from typing import Optional, Iterable, Iterator

from lxml import etree

from ..KML import AltitudeMode, State
from .Geometry import Geometry
from .LinearRing import LinearRing
from .Object import Object, ObjectChild


class Polygon(Geometry):
    """A Polygon geometry, as per https://developers.google.com/kml/documentation/kmlreference#polygon.
    :class:`~pyLiveKML.KML.KMLObjects.Polygon` objects are made up of an outer boundary that is a
    :class:`~pyLiveKML.KML.KMLObjects.LinearRing` and zero or more inner boundaries, each of which is also a
    :class:`~pyLiveKML.KML.KMLObjects.LinearRing`.

    :param LinearRing outer_boundary: A :class:`~pyLiveKML.KML.KMLObjects.LinearRing` that defines the outer extents of
        the :class:`~pyLiveKML.KML.KMLObjects.Polygon`.
    :param Optional[Iterable[LinearRing]] inner_boundaries: An (optional) iterable of
        :class:`~pyLiveKML.KML.KMLObjects.LinearRing` objects that define any cutouts within the
        :class:`~pyLiveKML.KML.KMLObjects.Polygon`.
    :param Optional[AltitudeMode] altitude_mode: The (optional) :class:`~pyLiveKML.KML.KML.AltitudeMode` that will be
        applied by GEP to all the points that make up the :class:`~pyLiveKML.KML.KMLObjects.Polygon`.
    :param Optional[bool] extrude: An (optional) flag to indicate whether the points that make up the
        :class:`~pyLiveKML.KML.KMLObjects.Polygon` should be shown in GEP connected to the ground with vertical lines.
    :param Optional[bool] tessellate: An (optional) flag to indicate whether the boundaries of the
        :class:`~pyLiveKML.KML.KMLObjects.Polygon` should follow the terrain.
    """

    @property
    def kml_type(self) -> str:
        """Overridden from :attr:`~pyLiveKML.KML.KMLObjects.Object.Object.kml_type` to set the KML tag name to
        'Polygon'"""
        return 'Polygon'

    @property
    def children(self) -> Iterator[ObjectChild]:
        """Overridden from :attr:`pyLiveKML.KML.KMLObjects.Object.Object.children` to yield the children of a
        :class:`~pyLiveKML.KML.KMLObjects.Polygon`, i.e. one or more :class:`~pyLiveKML.KML.KMLObjects.LinearRing`
        instances, being the :attr:`outer_boundary` and zero or more :attr:`inner_boundaries`.
        """
        yield ObjectChild(parent=self, child=self.outer_boundary)
        for b in self._inner_boundaries:
            yield ObjectChild(parent=self, child=b)

    @property
    def outer_boundary(self) -> LinearRing:
        """The :class:`~pyLiveKML.KML.KMLObjects.LinearRing` that defines the outer extents of the
        :class:`~pyLiveKML.KML.KMLObjects.Polygon`.
        """
        return self._outer_boundary

    @outer_boundary.setter
    def outer_boundary(self, value: LinearRing):
        if self._outer_boundary != value:
            self._outer_boundary = value
            self.field_changed()

    @property
    def inner_boundaries(self) -> Iterator[LinearRing]:
        """A generator to retrieve the :class:`~pyLiveKML.KML.KMLObjects.LinearRing` objects that define cutouts within
        the :attr:`outer_boundary`.

        :returns: A generator of :class:`~pyLiveKML.KML.KMLObjects.LinearRing` objects.
        """
        yield from self._inner_boundaries

    @property
    def extrude(self) -> Optional[bool]:
        """True if a vertical line (using the current :class:`~pyLiveKML.KML.KMLObjects.LineStyle`) connects each of
        the :attr:`outer_boundary` and :attr:`inner_boundaries` objects' points to the ground in GEP, False otherwise.
        None implies False.
        """
        return self._extrude

    @extrude.setter
    def extrude(self, value: Optional[bool]):
        if self._extrude != value:
            self._extrude = value
            self.field_changed()

    @property
    def tessellate(self) -> Optional[bool]:
        """True if the inner and outer boundary lines of the :class:`~pyLiveKML.KML.KMLObjects.Polygon` follows the
        terrain in GEP, otherwise False.

        :note: The :attr:`altitude_mode` property must be set to CLAMP_TO_GROUND to enable tessellation.
        """
        return self._tessellate

    @tessellate.setter
    def tessellate(self, value: Optional[bool]):
        if self._tessellate != value:
            self._tessellate = value
            self.field_changed()

    @property
    def altitude_mode(self) -> Optional[AltitudeMode]:
        """An :class:`~pyLiveKML.KML.KML.AltitudeMode` instance that defines how GEP displays the
        :class:`~pyLiveKML.KML.GeoCoordinates` objects that make up the inner and outer boundaries of the
        :class:`~pyLiveKML.KML.KMLObjects.Polygon` and treats their altitudes.
        """
        return self._altitude_mode

    @altitude_mode.setter
    def altitude_mode(self, value: Optional[AltitudeMode]):
        if self._altitude_mode != value:
            self._altitude_mode = value
            self.field_changed()

    def update_kml(self, parent: 'Object', update: etree.Element):
        # overrides the Object.update_kml() method to correctly handle the boundaries
        # Polygon boundaries are a special case for children, because they *must* be wrapped in an additional tag
        Object.update_kml(self, parent, update)
        if self._outer_boundary.state == State.CHANGING:
            self._outer_boundary.change_kml(update)
        self._outer_boundary.update_generated()
        for b in self._inner_boundaries:
            if b.state == State.CHANGING:
                b.change_kml(update)
            b.update_generated()

    def build_kml(self, root: etree.Element, with_children=True):
        if self._extrude is not None:
            etree.SubElement(root, 'extrude').text = str(int(self._extrude))
        if self._tessellate is not None:
            etree.SubElement(root, 'tessellate').text = str(int(self._tessellate))
        if self._altitude_mode is not None:
            etree.SubElement(root, 'altitudeMode').text = self._altitude_mode.value
        if with_children:
            if self._outer_boundary:
                etree.SubElement(root, 'outerBoundaryIs').append(self._outer_boundary.construct_kml())
                if self._outer_boundary._state == State.IDLE:
                    self._outer_boundary._state = State.CREATED
            for b in self._inner_boundaries:
                etree.SubElement(root, 'innerBoundaryIs').append(b.construct_kml())
                if b._state == State.IDLE:
                    b._state = State.CREATED

    def __init__(
            self,
            outer_boundary: LinearRing,
            inner_boundaries: Optional[Iterable[LinearRing]] = None,
            altitude_mode: Optional[AltitudeMode] = None,
            extrude: Optional[bool] = None,
            tessellate: Optional[bool] = None,
    ):
        Geometry.__init__(self)
        self._outer_boundary = outer_boundary
        self._inner_boundaries = list[LinearRing]()
        if inner_boundaries:
            self._inner_boundaries.extend(inner_boundaries)
        self._extrude = extrude
        self._tessellate = tessellate
        self._altitude_mode = altitude_mode



from abc import ABC, abstractmethod
from typing import Optional, NamedTuple, Iterator
from uuid import uuid4
from lxml import etree

from ..KML import State


class Object(ABC):
    """A KML 'Object', per https://developers.google.com/kml/documentation/kmlreference#object. Note that the
    :class:`~pyLiveKML.KML.KMLObjects.Object` class is explicitly abstract, and is the base class from which most other
    KML elements (anything with an :attr:`id` property) derive.
    """

    @property
    @abstractmethod
    def kml_type(self) -> str:
        """Abstract property that specifies the name of the XML tag that forms the root of the KML representation of
        this :class:`~pyLiveKML.KML.KMLObjects.Object`.
        """
        pass

    @property
    def id(self) -> uuid4:
        """The unique identifier of this :class:`~pyLiveKML.KML.KMLObjects.Object`.
        """
        return self._id

    @property
    def state(self) -> State:
        """The current GEP synchronization state of this :class:`~pyLiveKML.KML.KMLObjects.Object`.
        """
        return self._state

    @property
    def selected(self) -> bool:
        """True if this :class:`~pyLiveKML.KML.KMLObjects.Object` has been created and is not scheduled for deletion,
         otherwise False.
        """
        return (State.CREATING, State.CREATED, State.CHANGING).__contains__(self._state)

    @selected.setter
    def selected(self, value: bool):
        self.select(value)

    @property
    def children(self) -> Iterator['ObjectChild']:
        """A generator to retrieve the children of this :class:`~pyLiveKML.KML.KMLObjects.Object` as
        :class:`~pyLiveKML.KML.KMLObjects.Object.ObjectChild` instances.

        :returns: A generator of :class:`~pyLiveKML.KML.KMLObjects.Object.ObjectChild` named tuples that describes
            each child :class:`~pyLiveKML.KML.KMLObjects.Object` as a (parent, child)
        """
        # The return; yield pattern is used to trick PyCharm into accepting that this is a generator that yields
        # nothing, as opposed to yielding None
        return
        yield

    def build_kml(self, root: etree.Element, with_children: bool = True):
        """Generate the KML representation of the internal fields of this :class:`~pyLiveKML.KML.KMLObjects.Object`,
        and append it to the provided root etree.Element.

        :param etree.Element root: The root XML element that will be appended to.
        :param bool with_children: True if the children of this instance should be included in the build.
        """
        pass

    def construct_kml(self) -> etree.Element:
        """Constructs this :class:`~pyLiveKML.KML.KMLObjects.Object`'s KML representation.

        :returns: The KML representation of the object as an etree.Element.
        """
        root = etree.Element(_tag=self.kml_type, attrib={'id': str(self.id)})
        self.build_kml(root)
        return root

    def update_kml(self, parent: 'Object', update: etree.Element):
        """Retrieve a complete child <Create>, <Change> or <Delete> KML tag as a child of an <Update> tag.
        The type of child tag retrieved is dependent on the current state of this
        :class:`~pyLiveKML.KML.KMLObjects.Object`.

        :param Object parent: The immediate parent :class:`~pyLiveKML.KML.KMLObjects.Object` of this
            :class:`~pyLiveKML.KML.KMLObjects.Object`. The parent is required only for <Create> tags.
        :param etree.Element update: The etree.Element of the <Update> tag that will be appended to.
        """
        if self._state == State.CREATING:
            self.create_kml(parent, update)
        elif self._state == State.CHANGING:
            self.change_kml(update)
        elif self._state == State.DELETE_CREATED or self._state == State.DELETE_CHANGED:
            self.delete_kml(update)
        self.update_generated()

    def create_kml(self, parent: 'Object', update: etree.Element) -> etree.Element:
        """Construct a complete <Create> element tree as a child of an <Update> tag.

        :param Object parent: The immediate parent :class:`~pyLiveKML.KML.KMLObjects.Object` of this
            :class:`~pyLiveKML.KML.KMLObjects.Object`. The parent must be specified for GEP synchronization.
        :param etree.Element update: The etree.Element of the <Update> tag that will be appended to.
        """
        create = etree.Element('Create')
        parent_element = etree.SubElement(create, _tag=parent.kml_type, attrib={'targetId': str(parent.id)})
        item = self.construct_kml()
        parent_element.append(item)
        update.append(create)
        return item

    def change_kml(self, update: etree.Element):
        """Construct a complete <Change> element tree as a child of an <Update> tag.

        :param etree.Element update: The etree.Element of the <Update> tag that will be appended to.
        """
        change = etree.Element('Change')
        item = etree.SubElement(change, _tag=self.kml_type, attrib={'targetId': str(self.id)})
        self.build_kml(item, with_children=False)
        update.append(change)

    def delete_kml(self, update: etree.Element):
        """Construct a complete <Delete> element tree as a child of an <Update> tag.

        :param etree.Element update: The etree.Element of the <Update> tag that will be appended to.
        """
        delete = etree.Element('Delete')
        etree.SubElement(delete, _tag=self.kml_type, attrib={'targetId': str(self.id)})
        update.append(delete)

    def force_idle(self):
        """Force this :class:`~pyLiveKML.KML.KMLObjects.Object` and **all of its children** to the IDLE state.
        This is typically done after the object has been deselected, which will cause it to be deleted from GEP at the
        next synchronization update. When the :class:`~pyLiveKML.KML.KMLObjects.Object` is deleted by GEP, all of its
        children, and any :class:`~pyLiveKML.KML.KMLObjects.Feature` objects that it encloses, will also be
        automatically deleted. There is no need to emit <Delete> tags for these deletions, and in fact doing so will
        likely cause GEP to have conniptions.
        """
        self._state = State.IDLE
        for c in self.children:
            c.child.force_idle()

    def field_changed(self):
        """Flag that a field or property of this :class:`~pyLiveKML.KML.KMLObjects.Object` has changed, and
        re-synchronization with GEP may be required.
        """
        if self._state == State.CREATED:  # or self._state == State.IDLE:
            self._state = State.CHANGING
        elif self._state == State.DELETE_CREATED:
            self._state = State.DELETE_CHANGED
        elif self._state == State.IDLE:
            pass

    def update_generated(self):
        """Modify the state of the :class:`~pyLiveKML.KML.KMLObjects.Object` to reflect that a synchronization update
        has been emitted.
        """
        if self._state == State.CREATING:
            self._state = State.CREATED
            # if the object is being created, so are all of its descendants, in a single tag; set them created too
            for c in self.children:
                c.child.update_generated()
        elif self._state == State.CHANGING:
            # if the object is changing, don't mess with its descendants - they are updated elsewhere if necessary
            self._state = State.CREATED
        elif self._state == State.DELETE_CREATED or self._state == State.DELETE_CHANGED:
            self._state = State.IDLE

    def select(self, value: bool, cascade: bool = False):
        """Select or deselect this :class:`~pyLiveKML.KML.KMLObjects.Object` for display in GEP.

        :param bool value: True for selection, False for deselection
        :param bool cascade: True if the selection is to be cascaded to all child Objects.
        """
        if self._state == State.CREATING:
            self._state = State.IDLE if not value else self._state
        elif self._state == State.CREATED:
            self._state = State.DELETE_CREATED if not value else self._state
        elif self._state == State.CHANGING:
            self._state = State.DELETE_CHANGED if not value else self._state
        elif self._state == State.DELETE_CREATED:
            self._state = State.CREATING if value else self._state
        elif self._state == State.DELETE_CHANGED:
            self._state = State.CHANGING if value else self._state
        else:  # implies default state is IDLE
            self._state = State.CREATING if value else self._state
        # cascade Select downwards for Children
        if value:
            for c in self.children:
                c.child.select(True)
        else:
            for c in self.children:
                c.child.force_idle()

    def __init__(self):
        ABC.__init__(self)
        self._id: uuid4 = uuid4()
        self._selected: bool = False
        self._container: Optional[Object] = None
        self._state = State.IDLE

    def __str__(self):
        return f'{self.kml_type}'

    def __repr__(self):
        return self.__str__()


ObjectChild = NamedTuple('ObjectChild', [('parent', Object), ('child', Object)])
"""Named tuple that describes a parent:child relationship between two :class:`~pyLiveKML.KML.KMLObjects.Object` 
instances.
"""



from typing import Optional, Iterator

from lxml import etree

from .Object import ObjectChild
from .Style import Style
from .StyleSelector import StyleSelector


class StyleMap(StyleSelector):
    """A KML 'StyleMap', per https://developers.google.com/kml/documentation/kmlreference#stylemap. Maps between two
    different :class:`~pyLiveKML.KML.KMLObjects.Style` objects for the :attr:`~pyLiveKML.KML.KML.StyleState.NORMAL` and
    :attr:`~pyLiveKML.KML.KML.StyleState.HIGHLIGHT` states of a :class:`~pyLiveKML.KML.KMLObjects.Feature` when it is
    displayed in GEP. The child :class:`~pyLiveKML.KML.KMLObjects.Style` objects may be referenced by URI, or in-line
    in the :class:`~pyLiveKML.KML.KMLObjects.StyleMap`.

    :param Optional[str] normal_style_url: An (optional) URI reference for a :class:`~pyLiveKML.KML.KMLObjects.Style`
        to be employed in :attr:`~pyLiveKML.KML.KML.StyleState.NORMAL` display mode.
    :param Optional[Style] normal_style: An (optional) inline :class:`~pyLiveKML.KML.KMLObjects.Style` to be employed
        in :attr:`~pyLiveKML.KML.KML.StyleState.NORMAL` display mode.
    :param Optional[str] highlight_style_url: An (optional) URI reference for a
        :class:`~pyLiveKML.KML.KMLObjects.Style` to be employed in :attr:`~pyLiveKML.KML.KML.StyleState.HIGHLIGHT`
        display mode.
    :param Optional[Style] highlight_style: An (optional) inline :class:`~pyLiveKML.KML.KMLObjects.Style` to be
        employed in :attr:`~pyLiveKML.KML.KML.StyleState.HIGHLIGHT` display mode.

    :note: It is both possible and reasonable to assign both a :attr:`styleUrl` and a
        :class:`~pyLiveKML.KML.KMLObjects.Style` to either or both of the :attr:`~pyLiveKML.KML.KML.StyleState.NORMAL`
        and :attr:`~pyLiveKML.KML.KML.StyleState.HIGHLIGHT` states. Per the rules discussed at
        https://developers.google.com/kml/documentation/kmlreference#feature, an inline
        :class:`~pyLiveKML.KML.KMLObjects.Style` will override a URI-referenced
        :class:`~pyLiveKML.KML.KMLObjects.Style`.
    """

    @property
    def kml_type(self) -> str:
        """Overridden from :attr:`~pyLiveKML.KML.KMLObjects.Object.Object.kml_type` to set the KML tag name to
        'StyleMap'"""
        return 'StyleMap'

    @property
    def children(self) -> Iterator[ObjectChild]:
        """Overridden from :attr:`pyLiveKML.KML.KMLObjects.Object.Object.children` to yield the children of a
        :class:`~pyLiveKML.KML.KMLObjects.Style`, i.e. up to two :class:`~pyLiveKML.KML.KMLObjects.Style` instances.
        """
        if self._normal_style:
            yield ObjectChild(parent=self, child=self._normal_style)
        if self._highlight_style:
            yield ObjectChild(parent=self, child=self._highlight_style)

    @property
    def normal_style_url(self) -> Optional[str]:
        """A URI that references the :class:`~pyLiveKML.KML.KMLObjects.Style` that will be employed when the target
        :class:`~pyLiveKML.KML.KMLObjects.Feature` is displayed in :attr:`~pyLiveKML.KML.KML.StyleState.NORMAL` mode.
        """
        return self._normal_style_url
    
    @normal_style_url.setter
    def normal_style_url(self, value: Optional[str]):
        if self._normal_style_url != value:
            self._normal_style_url = value
            self.field_changed()

    @property
    def normal_style(self) -> Optional[Style]:
        """An inline :class:`~pyLiveKML.KML.KMLObjects.Style` that will be employed when the target
        :class:`~pyLiveKML.KML.KMLObjects.Feature` is displayed in :attr:`~pyLiveKML.KML.KML.StyleState.NORMAL` mode.
        """
        return self._normal_style

    @normal_style.setter
    def normal_style(self, value: Optional[Style]):
        if self._normal_style != value:
            self._normal_style = value
            self.field_changed()

    @property
    def highlight_style_url(self) -> Optional[str]:
        """A URI that references the :class:`~pyLiveKML.KML.KMLObjects.Style` that will be employed when the target
        :class:`~pyLiveKML.KML.KMLObjects.Feature` is displayed in :attr:`~pyLiveKML.KML.KML.StyleState.HIGHLIGHT` mode.
        """
        return self._highlight_style_url

    @highlight_style_url.setter
    def highlight_style_url(self, value: Optional[str]):
        if self._highlight_style_url != value:
            self._highlight_style_url = value
            self.field_changed()

    @property
    def highlight_style(self) -> Optional[Style]:
        """An inline :class:`~pyLiveKML.KML.KMLObjects.Style` that will be employed when the target
        :class:`~pyLiveKML.KML.KMLObjects.Feature` is displayed in :attr:`~pyLiveKML.KML.KML.StyleState.HIGHLIGHT` mode.
        """
        return self._highlight_style

    @highlight_style.setter
    def highlight_style(self, value: Optional[Style]):
        if self._highlight_style != value:
            self._highlight_style = value
            self.field_changed()

    def build_kml(self, root: etree.Element, with_children=True):
        if with_children:
            if self._normal_style_url or self._normal_style:
                normal = etree.SubElement(root, 'Pair')
                etree.SubElement(normal, 'key').text = 'normal'
                if self._normal_style:
                    normal.append(self._normal_style.construct_kml())
                if self._normal_style_url:
                    etree.SubElement(normal, 'styleUrl').text = self._normal_style_url
            if self._highlight_style_url or self._highlight_style:
                highlight = etree.SubElement(root, 'Pair')
                etree.SubElement(highlight, 'key').text = 'highlight'
                if self._highlight_style:
                    highlight.append(self._highlight_style.construct_kml())
                if self._highlight_style_url:
                    etree.SubElement(highlight, 'styleUrl').text = self._highlight_style_url

    def __init__(
            self,
            normal_style_url: Optional[str] = None,
            normal_style: Optional[Style] = None,
            highlight_style_url: Optional[str] = None,
            highlight_style: Optional[Style] = None,
    ):
        super().__init__()
        self._normal_style_url = normal_style_url
        self._normal_style = normal_style
        self._highlight_style_url = highlight_style_url
        self._highlight_style = highlight_style
            


from abc import ABC

from .Object import Object


class StyleSelector(Object, ABC):
    """A KML 'StyleSelector', per https://developers.google.com/kml/documentation/kmlreference#styleselector. The
    :class:`~pyLiveKML.KML.KMLObjects.StyleSelector` class is the abstract base class for KML
    :class:`~pyLiveKML.KML.KMLObjects.Object` instances that represent display styles for
    :class:`~pyLiveKML.KML.KMLObjects.Feature` instances.
    """
    def __init__(self):
        Object.__init__(self)
        ABC.__init__(self)



from abc import ABC
from typing import Optional, Iterable, Iterator

from ..KML import State
from ..KMLObjects.Object import Object
from ..KMLObjects.StyleSelector import StyleSelector


class Feature(Object, ABC):
    """A KML 'Feature', per https://developers.google.com/kml/documentation/kmlreference#feature. Note that while
    Features are explicitly abstract in the KML specification, :class:`~pyLiveKML.KML.KMLObjects.Feature` is the base
    class for KML :class:`~pyLiveKML.KML.KMLObjects.Object` instances that have an "existence" in GEP, i.e. that are
    (potentially) user-editable because they appear in the GEP user List View.

    :param Optional[str] name: The (optional) name for this :class:`~pyLiveKML.KML.KMLObjects.Feature` that will be
        displayed in GEP.
    :param Optional[str] description: The (optional) description for this :class:`~pyLiveKML.KML.KMLObjects.Feature`
        that will be displayed in GEP as a text balloon if the :class:`~pyLiveKML.KML.KMLObjects.Feature` is clicked.
    :param Optional[bool] visibility: The (optional) initial visibility for this
        :class:`~pyLiveKML.KML.KMLObjects.Feature` in GEP.
    :param Optional[Feature] container: The (optional) :class:`~pyLiveKML.KML.KMLObjects.Feature` (generally, a
        :class:`~pyLiveKML.KML.KMLObjects.Container`) that encloses this
        :class:`~pyLiveKML.KML.KMLObjects.Feature`.
    :param Optional[str] style_url: An (optional) style URL, typically a reference to a global
        :class:`~pyLiveKML.KML.KMLObjects.StyleSelector` in a :class:`~pyLiveKML.KML.KMLObjects.Container` that
        encloses this :class:`~pyLiveKML.KML.KMLObjects.Feature`.
    :param Optional[Iterable[StyleSelector]] styles: An iterable of :class:`~pyLiveKML.KML.KMLObjects.StyleSelector`
        objects that are local to this :class:`~pyLiveKML.KML.KMLObjects.Feature`.
    """

    @property
    def container(self) -> Optional['Container']:
        """The :class:`~pyLiveKML.KML.KMLObjects.Container` that immediately encloses this
        :class:`~pyLiveKML.KML.KMLObjects.Feature` in an ownership tree.

        :warning: The :attr:`container` property cannot be altered if the :class:`~pyLiveKML.KML.KMLObjects.Feature` is
            visible in GEP. Doing so would break GEP synchronization. Failure to observe this constraint will cause a
            :class:`ValueError` to be raised.
        """
        return self._container

    @container.setter
    def container(self, value: 'Container'):
        if self._state == State.IDLE or self._state == State.CREATING:
            self._container = value
        else:
            raise ValueError('If a Feature is visible in GEP, you cannot change its\' \'container\' property.')

    @property
    def name(self) -> Optional[str]:
        """The text that will be displayed as the name of the :class:`~pyLiveKML.KML.KMLObjects.Feature` in GEP.
        """
        return self._name

    @name.setter
    def name(self, value: Optional[str]):
        if self._name != value:
            self._name = value
            self.field_changed()

    @property
    def visibility(self) -> Optional[bool]:
        """True if the :class:`~pyLiveKML.KML.KMLObjects.Feature` will (initially) be checked (visible) in GEP, False
        otherwise.
        """
        return self._visibility

    @visibility.setter
    def visibility(self, value: Optional[bool]):
        if self._visibility != value:
            self._visibility = value
            self.field_changed()

    @property
    def description(self) -> Optional[str]:
        """The text description for this :class:`~pyLiveKML.KML.KMLObjects.Feature`, that will be displayed in a
        balloon in GEP if the :class:`~pyLiveKML.KML.KMLObjects.Feature` is clicked.

        :note: HTML and (some) JavaScript are permissible (in GEP > 5.0) for the :attr:`description` property. Refer to
            the KML specification at https://developers.google.com/kml/documentation/kmlreference for details.
        """
        return self._description

    @description.setter
    def description(self, value: Optional[str]):
        if self._description != value:
            self._description = value
            self.field_changed()

    @property
    def styles(self) -> Iterator[StyleSelector]:
        """A generator to retrieve references to any :class:`~pyLiveKML.KML.KMLObjects.Style` or
        :class:`~pyLiveKML.KML.KMLObjects.StyleMap` objects that are children of this
        :class:`~pyLiveKML.KML.KMLObjects.Feature`.

        :returns: A generator of :class:`~pyLiveKML.KML.KMLObjects.StyleSelector` objects.
        """
        for s in self._styles:
            yield s

    # override Object.select() to enable upwards cascade, i.e. if a Feature contained in an unselected parent Feature
    # is selected, the parent Feature must also be selected in order for GEP synchronization to work correctly.
    def select(self, value: bool, cascade: bool = False):
        """Overrides :func:`~pyLiveKML.KML.KMLObjects.Object.Object.select` to implement upwards cascade of selection.
        That is, if a :class:`~pyLiveKML.KML.KMLObjects.Feature` enclosed in the object tree depending from an
        unselected  parent :class:`~pyLiveKML.KML.KMLObjects.Feature` is selected, the reverse tree's parents must also
        be selected in order for GEP synchronization to work correctly.
        """
        Object.select(self, value, cascade)
        # Cascade Select *upwards* for Features, but *do not* cascade Deselect upwards
        if value and self._container:
            self._container.select(True, False)

    def __init__(
            self,
            name: Optional[str] = None,
            description: Optional[str] = None,
            visibility: Optional[bool] = None,
            container: Optional['Feature'] = None,
            style_url: Optional[str] = None,
            styles: Optional[Iterable[StyleSelector]] = None,
    ):
        Object.__init__(self)
        ABC.__init__(self)
        self._container = container
        self._name = name
        self._visibility = visibility
        self._description = description
        self._style_url = style_url
        self._styles = list[StyleSelector]()
        if styles:
            self._styles.extend(styles)

    def __str__(self):
        return f'{self.kml_type}:{self.name}'

    def __repr__(self):
        return self.__str__()



from typing import Optional

from lxml import etree

from .ColorStyle import ColorStyle


class LineStyle(ColorStyle):
    """
    A KML 'LineStyle', per https://developers.google.com/kml/documentation/kmlreference#linestyle.  Specifies
    various properties of a :class:`~pyLiveKML.KML.KMLObjects.Geometry`, typically either a
    :class:`~pyLiveKML.KML.KMLObjects.LineString` or :class:`~pyLiveKML.KML.KMLObjects.LinearRing`, including
    the inner and outer boundaries of :class:`~pyLiveKML.KML.KMLObjects.Polygon` objects.

    :param Optional[float] width: The (optional) width of the line, in pixels.
    :param Optional[int] color: The (optional) color of the line, as a 32-bit ABGR value.
    """

    @property
    def kml_type(self) -> str:
        """Overridden from :attr:`~pyLiveKML.KML.KMLObjects.Object.Object.kml_type` to set the KML tag name to
        'LineStyle'"""
        return 'LineStyle'

    @property
    def width(self) -> Optional[float]:
        """The width of the line, in pixels.
        """
        return self._width

    @width.setter
    def width(self, value: Optional[float]):
        if self._width != value:
            self._width = value
            self.field_changed()

    @property
    def gx_outer_color(self) -> Optional[int]:
        """Color of the portion of the line defined by the :attr:`gx_outer_width` property. Note that both
        :attr:`gx_outer_color` and :attr:`gx_outer_width` apply only to :class:`~pyLiveKML.KML.KMLObjects.LineString`
        objects.
        """
        return self._gx_outer_color

    @gx_outer_color.setter
    def gx_outer_color(self, value: Optional[int]):
        val = 0 if value <= 0 else 0xffffffff if value >= 0xffffffff else value
        if self._gx_outer_color != val:
            self._gx_outer_color = val
            self.field_changed()

    @property
    def gx_outer_width(self) -> Optional[float]:
        """Width of an outer border around the line. Note that both :attr:`gx_outer_color` and :attr:`gx_outer_width`
        apply only to :class:`~pyLiveKML.KML.KMLObjects.LineString` objects.
        """
        return self._gx_outer_width

    @gx_outer_width.setter
    def gx_outer_width(self, value: Optional[float]):
        if self._gx_outer_width != value:
            self._gx_outer_width = value
            self.field_changed()

    @property
    def gx_physical_width(self) -> Optional[float]:
        """The physical width of the line, in metres.
        """
        return self._gx_physical_width

    @gx_physical_width.setter
    def gx_physical_width(self, value: Optional[float]):
        if self._gx_physical_width != value:
            self._gx_physical_width = value
            self.field_changed()

    @property
    def gx_label_visibility(self) -> Optional[bool]:
        """Selector to display a text label on a :class:`~pyLiveKML.KML.KMLObjects.LineString`. True if the label is to
        be displayed, False if it is not. None implies False.
        """
        return self._gx_label_visibility

    @gx_label_visibility.setter
    def gx_label_visibility(self, value: Optional[bool]):
        if self._gx_label_visibility != value:
            self._gx_label_visibility = value
            self.field_changed()

    def build_kml(self, root: etree.Element, with_children=True):
        if self._color is not None:
            etree.SubElement(root, 'color').text = f'{self._color:08x}'
        if self._color_mode is not None:
            etree.SubElement(root, 'colorMode').text = self._color_mode.value
        if self._width is not None:
            etree.SubElement(root, 'width').text = f'{self._width:0.1f}'
        if self._gx_outer_color is not None:
            etree.SubElement(root, 'gx:outerColor').text = f'{self._gx_outer_color:08x}'
        if self._gx_outer_width is not None:
            etree.SubElement(root, 'gx:outerWidth').text = f'{self._gx_outer_width:0.1f}'
        if self._gx_physical_width is not None:
            etree.SubElement(root, 'gx:physicalWidth').text = f'{self._gx_physical_width:0.1f}'
        if self._gx_label_visibility is not None:
            etree.SubElement(root, 'gx:labelVisibility').text = str(int(self._gx_label_visibility))

    def __init__(
            self,
            width: Optional[float] = None,
            color: Optional[int] = None,
    ):
        ColorStyle.__init__(self, color)
        self._width = width
        self._gx_outer_color: Optional[int] = None
        self._gx_outer_width: Optional[float] = None
        self._gx_physical_width: Optional[float] = None
        self._gx_label_visibility: Optional[bool] = None

    def __str__(self):
        return f'{self.kml_type}'

    def __repr__(self):
        return self.__str__()



from typing import Optional

from lxml import etree

from ..KML import ColorMode
from .ColorStyle import ColorStyle
from ..Vec2 import Vec2


class IconStyle(ColorStyle):
    """
    A KML 'IconStyle', per https://developers.google.com/kml/documentation/kmlreference#iconstyle.  Specifies
    various properties that define how a :class:`~pyLiveKML.KML.KMLObjects.Icon` is drawn.  Applies to
    :class:`~pyLiveKML.KML.KMLObjects.Point` geometries.

    :param str icon: A URI for an image or icon file.
    :param float scale: The (optional) relative scale of the icon.
    :param float heading: The (optional) heading, in degrees, that the icon will be rotated to point towards.
    :param Optional[int] color: The (optional) color of the icon, as a 32-bit ABGR value.
    :param Optional[ColorMode] color_mode: The (optional) :class:`~pyLiveKML.KML.KML.ColorMode` used to color the icon;
        either 'NORMAL' or 'RANDOM'.
    """

    @property
    def kml_type(self) -> str:
        """Overridden from :attr:`~pyLiveKML.KML.KMLObjects.Object.Object.kml_type` to set the KML tag name to
        'IconStyle'"""
        return 'IconStyle'

    @property
    def icon(self) -> Optional[str]:
        """URI for the image that will be displayed in GEP as the icon.
        """
        return self._icon

    @icon.setter
    def icon(self, value: Optional[str]):
        if self._icon != value:
            self._icon = value
            self.field_changed()

    @property
    def scale(self) -> Optional[float]:
        """Relative scale at which the icon will be displayed in GEP.
        """
        return self._scale

    @scale.setter
    def scale(self, value: Optional[float]):
        if self._scale != value:
            self._scale = value
            self.field_changed()

    @property
    def heading(self) -> Optional[float]:
        """Heading (in degrees) that the icon will be displayed pointing towards in GEP.
        """
        return self._heading

    @heading.setter
    def heading(self, value: Optional[float]):
        if self._heading != value:
            self._heading = value
            self.field_changed()

    @property
    def hotspot(self) -> Optional[Vec2]:
        """Relative position in the icon that is anchored to the associated :class:`~pyLiveKML.KML.KMLObjects.Point`
        """
        return self._hotspot

    @hotspot.setter
    def hotspot(self, value: Optional[Vec2]):
        value.name = 'hotSpot'
        if self._hotspot != value:
            self._hotspot = value
            self.field_changed()

    def build_kml(self, root: etree.Element, with_children=True):
        if self.color is not None:
            etree.SubElement(root, 'color').text = f'{self.color:08x}'
        if self.color_mode is not None:
            etree.SubElement(root, 'colorMode').text = self.color_mode.value
        if self.scale is not None:
            etree.SubElement(root, 'scale').text = f'{self.scale:0.3f}'
        if self.heading is not None:
            etree.SubElement(root, 'heading').text = f'{self.heading:0.1f}'
        if self.icon is not None:
            etree.SubElement(etree.SubElement(root, 'Icon'), 'href').text = self.icon
        if self.hotspot is not None:
            root.append(self.hotspot.xml)

    def __init__(
            self,
            icon: str,
            scale: float = 1.0,
            heading: float = 0.0,
            color: Optional[int] = None,
            color_mode: Optional[ColorMode] = None,
    ):
        ColorStyle.__init__(self, color=color, color_mode=color_mode)
        self._scale = scale
        self._heading = heading
        self._icon = icon
        self._hotspot: Optional[Vec2] = None

    def __str__(self):
        return f'{self.kml_type}'

    def __repr__(self):
        return self.__str__()



from abc import ABC
from typing import Optional, Iterable, NamedTuple, Iterator

from lxml import etree

from ..KML import KML_UPDATE_CONTAINER_LIMIT_DEFAULT, State
from .Feature import Feature
from .Object import Object, ObjectChild
from .StyleSelector import StyleSelector


class Container(list[Feature], Feature, ABC):
    """A KML 'Container', per https://developers.google.com/kml/documentation/kmlreference#container. Note that while
    Containers are explicitly abstract, :class:`~pyLiveKML.KML.KMLObjects.Container` is the base class for KML
    :class:`~pyLiveKML.KML.KMLObjects.Folder` and :class:`~pyLiveKML.KML.KMLObjects.Document` that have an "existence"
    in GEP, i.e. that are (potentially) user-editable because they appear in the GEP user List View, and that may
    'contain' other :class:`~pyLiveKML.KML.KMLObjects.Feature` objects, including other concrete
    :class:`~pyLiveKML.KML.KMLObjects.Container` objects.

    :param Optional[str] name: The (optional) name for this :class:`~pyLiveKML.KML.KMLObjects.Container` that will
        be displayed in GEP.
    :param Optional[bool] visibility: The (optional) initial visibility for this
        :class:`~pyLiveKML.KML.KMLObjects.Container` in GEP.
    :param Optional[bool] is_open: Optional boolean flag to indicate whether the
        :class:`~pyLiveKML.KML.KMLObjects.Container` will be displayed as 'open' in the GEP user List View.
    :param Optional[int] update_limit: Only applies to the root of a :class:`~pyLiveKML.KML.KMLObjects.Feature` tree.
        The (approximate) maximum number of contained :class:`~pyLiveKML.KML.KMLObjects.Feature` instances that will be
        synchronized from this :class:`~pyLiveKML.KML.KMLObjects.Container` during any one synchronization update.
    :param Optional[str] style_url: An (optional) style URL, generally a reference to a global
        :class:`~pyLiveKML.KML.KMLObjects.StyleSelector` in a parent of this
        :class:`~pyLiveKML.KML.KMLObjects.Container`.
    :param Optional[Iterable[StyleSelector]] styles: An (optional) Iterable of
        :class:`~pyLiveKML.KML.KMLObjects.StyleSelector` objects that will be local to this
        :class:`~pyLiveKML.KML.KMLObjects.Container`.
    :param Optional[Iterable[Feature]] features: An (optional) Iterable of :class:`~pyLiveKML.KML.KMLObjects.Feature`
        objects to be enclosed by this :class:`~pyLiveKML.KML.KMLObjects.Container`.
    """

    @property
    def containers(self) -> Iterator['ContainedFeature']:
        """A generator to retrieve references to any :class:`~pyLiveKML.KML.KMLObjects.Container` objects that are
        enclosed by this :class:`~pyLiveKML.KML.KMLObjects.Container` object, and the tree that is rooted at it.

        :returns: A generator of :class:`~pyLiveKML.KML.KMLObjects.Container.ContainedFeature` named tuples that
            describe each enclosed :class:`~pyLiveKML.KML.KMLObjects.Container` as a (container, feature)
        """
        for f in self:
            if isinstance(f, Container):
                yield ContainedFeature(container=self, feature=f)
                yield from f.containers

    @property
    def features(self) -> Iterator['ContainedFeature']:
        """A generator to retrieve references to the :class:`~pyLiveKML.KML.KMLObjects.Feature` objects that are
        enclosed by this :class:`~pyLiveKML.KML.KMLObjects.Container` object, and the tree that is rooted at it. Note
        that :class:`~pyLiveKML.KML.KMLObjects.Container` objects are *not* yielded by this generator, despite being
        specializations of :class:`~pyLiveKML.KML.KMLObjects.Feature`; use the :attr:`containers` property to retrieve
        them.

        :returns: A generator of :class:`~pyLiveKML.KML.KMLObjects.Container.ContainedFeature` named tuples that
            describes each enclosed :class:`~pyLiveKML.KML.KMLObjects.Feature` as a (container, feature)
        """
        for f in self:
            if isinstance(f, Container):
                yield from f.features
            elif isinstance(f, Feature):
                yield ContainedFeature(container=self, feature=f)

    @property
    def children(self) -> Iterator[ObjectChild]:
        """Overridden from :attr:`pyLiveKML.KML.KMLObjects.Object.Object.children` to yield the children of a
        :class:`~pyLiveKML.KML.KMLObjects.Container`, i.e. one or more :class:`~pyLiveKML.KML.KMLObjects.StyleSelector`
        instances, and their children.
        """
        for s in self._styles:
            yield ObjectChild(parent=self, child=s)
            yield from s.children

    @property
    def is_open(self) -> Optional[bool]:
        """True if the :class:`~pyLiveKML.KML.KMLObjects.Container` will be initially displayed in an 'open' state in
        the GEP user List View, else False if it will be initially displayed in a 'closed' state.  None implies the
        default of False.
        """
        return self._is_open

    @is_open.setter
    def is_open(self, value: Optional[bool]):
        if self._is_open != value:
            self._is_open = value
            self.field_changed()

    @property
    def update_limit(self) -> int:
        """The (approximate) maximum number of KML objects that will be synchronized by any single
        synchronization update that is rooted in this :class:`~pyLiveKML.KML.KMLObjects.Container`.
        """
        return self._update_limit

    @update_limit.setter
    def update_limit(self, value: int):
        # note that because this is not a KML field, i.e. that is displayed by GEP, there is no need to call
        # self.field_changed() if the value is updated
        if value != self._update_limit and value > 0:
            self._update_limit = value

    @property
    def flush(self) -> Iterator[Feature]:
        """A generator to retrieve instances of :class:`~pyLiveKML.KML.KMLObjects.Feature` objects that have been
        deleted from this :class:`~pyLiveKML.KML.KMLObjects.Container` but for which those deletions **have not** yet
        been synchronized with GEP. As the generator retrieves a deleted :class:`~pyLiveKML.KML.KMLObjects.Feature`, it
        also completes the deletion process.

        :returns: A generator of :class:`~pyLiveKML.KML.KMLObjects.Feature` objects.
        """
        while len(self.__deleted) > 0:
            f = self.__deleted[0]
            self.__deleted.remove(f)
            yield f

    def construct_kml(self, with_features: bool = False) -> etree.Element:
        """Overridden from :func:`~pyLiveKML.KML.KMLObjects.Object.construct_kml` to allow for the creation of
        contained or enclosed :class:`~pyLiveKML.KML.KMLObjects.Feature` instances, including other
        :class:`~pyLiveKML.KML.KMLObjects.Container` instances.
        """
        root = Object.construct_kml(self)
        if with_features:
            for f in self:
                if isinstance(f, Container):
                    root.append(f.construct_kml(with_features=True))
                elif isinstance(f, Feature):
                    root.append(f.construct_kml())
        return root

    def build_kml(self, root: etree.Element, with_children=True):
        if self.name is not None:
            etree.SubElement(root, 'name').text = self.name
        if self.visibility is not None:
            etree.SubElement(root, 'visibility').text = str(int(self.visibility))
        if self.is_open is not None:
            etree.SubElement(root, 'open').text = str(int(self.is_open))
        if self.description is not None:
            etree.SubElement(root, 'description').text = self.description
        if with_children:
            for s in self._styles:
                root.append(s.construct_kml())

    def append(self, item: Feature):
        """Append a :class:`~pyLiveKML.KML.KMLObjects.Feature` to this :class:`~pyLiveKML.KML.KMLObjects.Container`.

        :param Feature item: The :class:`~pyLiveKML.KML.KMLObjects.Feature` to be appended.
        """
        list[Feature].append(self, item)
        item.container = self

    def remove(self, __value: Feature) -> None:
        """Remove a :class:`~pyLiveKML.KML.KMLObjects.Feature` from this :class:`~pyLiveKML.KML.KMLObjects.Container`.
        Of course, the :class:`~pyLiveKML.KML.KMLObjects.Feature` must be enclosed in this
        :class:`~pyLiveKML.KML.KMLObjects.Container` to be able to be removed.

        :param Feature __value: The :class:`~pyLiveKML.KML.KMLObjects.Feature` to be removed.
        """
        if __value.selected:
            self.__deleted.append(__value)
        list[Feature].remove(self, __value)

    def force_idle(self, cascade: bool = False):
        """Overridden from :func:`~pyLiveKML.KML.KMLObjects.Object.Object.force_idle` to enable the entire tree of
        enclosed :class:`~pyLiveKML.KML.KMLObjects.Feature` (and :class:`~pyLiveKML.KML.KMLObjects.Container`)
        instances, and child :class:`~pyLiveKML.KML.KMLObjects.Object` instances, that is rooted in this
        :class:`~pyLiveKML.KML.KMLObjects.Container` to be forced to the IDLE state.
        """
        Object.force_idle(self)
        if cascade:
            self.force_features_idle()

    def force_features_idle(self):
        """Force the entire tree of enclosed :class:`~pyLiveKML.KML.KMLObjects.Feature` (and
        :class:`~pyLiveKML.KML.KMLObjects.Container`) instances to be forced to the IDLE state. Typically called as a
        result of the target :class:`~pyLiveKML.KML.KMLObjects.Container` being deleted from GEP.
        """
        for f in self:
            if isinstance(f, Container):
                # note the implication from force_idle() that cascade is _always_ true for force_features_idle
                f.force_idle(True)
            elif isinstance(f, Feature):
                f.force_idle()

    def select(self, value: bool, cascade: bool = False):
        """Overrides :func:`~pyLiveKML.KML.KMLObjects.Feature.Feature.select` to implement select/deselect cascade to
        enclosed :class:`~pyLiveKML.KML.KMLObjects.Feature` objects, and to ensure that if a
        :class:`~pyLiveKML.KML.KMLObjects.Container` is deleted from GEP, its' enclosed
        :class:`~pyLiveKML.KML.KMLObjects.Feature` objects are forced IDLE to maintain synchronization.
        """
        Feature.select(self, value, cascade)
        if cascade:
            for f in self:
                f.select(value, True)
        if self._state == State.DELETE_CREATED or self._state == State.DELETE_CHANGED:
            self.__deleted.clear()
            self.force_features_idle()

    def __init__(
            self,
            name: Optional[str] = None,
            visibility: Optional[bool] = None,
            is_open: Optional[bool] = None,
            update_limit: int = KML_UPDATE_CONTAINER_LIMIT_DEFAULT,
            style_url: Optional[str] = None,
            styles: Optional[Iterable[StyleSelector]] = None,
            features: Optional[Iterable[Feature]] = None,
    ):
        list[Feature].__init__(self)
        Feature.__init__(
            self,
            name=name,
            visibility=visibility,
            style_url=style_url,
            styles=styles
        )
        ABC.__init__(self)
        if features:
            self.extend(features)
        self._is_open = is_open
        self._update_limit = KML_UPDATE_CONTAINER_LIMIT_DEFAULT
        self.update_limit = update_limit
        self.__deleted = list[Feature]()

    def __str__(self):
        return Feature.__str__(self)

    def __repr__(self):
        return Feature.__repr__(self)


ContainedFeature = NamedTuple('ContainedFeature', [('container', Container), ('feature', Feature)])
"""Named tuple that describes a container:contained relationship between a :class:`~pyLiveKML.KML.KMLObjects.Container` 
instance and a :class:`~pyLiveKML.KML.KMLObjects.Feature` instance.
"""



from typing import Optional, Iterator

from lxml import etree

from ..KML import RefreshMode
from .Feature import Feature
from .Link import Link
from .Object import ObjectChild


class NetworkLink(Feature):
    """A KML 'NetworkLink', per https://developers.google.com/kml/documentation/kmlreference#networklink.
    :class:`~pyLiveKML.KML.KMLObjects.NetworkLink` objects are typically used to direct GEP to periodically retrieve a
    file from a specified href.

    :param Optional[str] name: The (optional) text that will be displayed in the GEP user List View as the name of the
        :class:`~pyLiveKML.KML.KMLObjects.NetworkLink`.
    :param Optional[str] href: An (optional) href for the file that will be loaded by the
        :class:`~pyLiveKML.KML.KMLObjects.NetworkLink`.
    :param Optional[RefreshMode] refresh_mode: The (optional) refresh mode that will be used for file loading.
    :param Optional[float] refresh_interval: The (optional) refresh interval, in seconds, that will be used for file
        loading.
    :param Optional[bool] is_open: Optional flag to indicate whether the :class:`~pyLiveKML.KML.KMLObjects.NetworkLink`
        will be displayed as 'open' in the GEP user List View.
    """

    @property
    def kml_type(self) -> str:
        """Overridden from :attr:`~pyLiveKML.KML.KMLObjects.Object.Object.kml_type` to set the KML tag name to
        'NetworkLink'"""
        return 'NetworkLink'

    @property
    def children(self) -> Iterator[ObjectChild]:
        if self._link:
            yield self._link
        yield from self.styles

    @property
    def is_open(self) -> Optional[bool]:
        """True if the :class:`~pyLiveKML.KML.KMLObjects.NetworkLink` will be initially displayed in an 'open' state
        in the GEP user List View, else False if it will be initially displayed in a 'closed' state.  None implies the
        default of False.
        """
        return self._is_open

    @is_open.setter
    def is_open(self, value: Optional[bool]):
        if self._is_open != value:
            self._is_open = value
            self.field_changed()

    @property
    def link(self) -> Link:
        """The child :class:`~pyLiveKML.KML.KMLObjects.Link` object that identifies how and from where this
        :class:`~pyLiveKML.KML.KMLObjects.NetworkLink` will load its dependent file.
        """
        return self._link

    @property
    def fly_to_view(self) -> Optional[bool]:
        """True if this :class:`~pyLiveKML.KML.KMLObjects.NetworkLink` instructs GEP to fly to its view location when
        loaded, else False if it does not. None implies the default of False.
        """
        return self._fly_to_view

    @fly_to_view.setter
    def fly_to_view(self, value: Optional[bool]):
        if self._fly_to_view != value:
            self._fly_to_view = value

    @property
    def refresh_visibility(self) -> Optional[bool]:
        """True if the GEP user is not permitted to control the visibility of the
        :class:`~pyLiveKML.KML.KMLObjects.NetworkLink` or its children, else False if the GEP user has full control
        over that visibility.  None implies the default of False.
        """
        return self._refresh_visibility

    @refresh_visibility.setter
    def refresh_visibility(self, value: Optional[bool]):
        if self._refresh_visibility != value:
            self._refresh_visibility = value

    def build_kml(self, root: etree.Element, with_children=True):
        if self._name:
            etree.SubElement(root, 'name').text = self._name
        if self._visibility is not None:
            etree.SubElement(root, 'visibility').text = str(int(self._visibility))
        if self._is_open is not None:
            etree.SubElement(root, 'open').text = str(int(self._is_open))
        if self._description:
            etree.SubElement(root, 'description').text = self._description
        if self._refresh_visibility is not None:
            etree.SubElement(root, 'refreshVisibility').text = str(int(self._refresh_visibility))
        if self._style_url:
            etree.SubElement(root, 'styleUrl').text = self._style_url
        if with_children:
            for s in self._styles:
                root.append(s.construct_kml())
            if self._link:
                root.append(self._link.construct_kml())

    def __init__(
            self,
            name: Optional[str] = None,
            href: Optional[str] = None,
            refresh_mode: Optional[RefreshMode] = None,
            refresh_interval: Optional[float] = None,
            is_open: Optional[bool] = None,
    ):
        Feature.__init__(self, name=name, visibility=None)
        self._is_open = is_open
        self._link = Link(href, refresh_mode, refresh_interval)
        self._fly_to_view = None
        self._refresh_visibility = None



from typing import Optional, NamedTuple

from lxml import etree

from ..KML import RefreshMode
from .Link import Link


GxParams = NamedTuple('GxParams', [('x', int), ('y', int), ('w', int), ('h', int)])


class Icon(Link):
    """A KML 'Icon', per https://developers.google.com/kml/documentation/kmlreference#icon.
    :class:`~pyLiveKML.KML.KMLObjects.Icon` instances are used to specify an image file that will be displayed in a GEP
    overlay.  Note that only time-based refresh of :class:`~pyLiveKML.KML.KMLObjects.Icon` instances is currently
    supported by pyLiveKML.

    :param Optional[str] href: An (optional) href for the file that is referenced by the
        :class:`~pyLiveKML.KML.KMLObjects.Icon`.
    :param Optional[RefreshMode] refresh_mode: The (optional) :class:`~pyLiveKML.KML.KML.RefreshMode` that will be used
        for file loading.
    :param Optional[float] refresh_interval: The (optional) refresh interval, in seconds, that will be used for file
        loading.
    :param Optional[GxParams] gx_params: Optional :class:`GxParams` instance that defines how GEP will treat the icon
        image.
    """

    @property
    def kml_type(self) -> str:
        """Overridden from :attr:`~pyLiveKML.KML.KMLObjects.Object.Object.kml_type` to set the KML tag name to
        'Icon'"""
        return 'Icon'

    @property
    def gx_params(self) -> Optional[GxParams]:
        """A :class:`GxParams` named tuple that describes how the icon is treated by GEP.
        """
        return self._gx_params

    @gx_params.setter
    def gx_params(self, value: Optional[GxParams]):
        self._gx_params = value
        self.field_changed()

    def build_kml(self, root: etree.Element, with_children=True):
        if self._href:
            etree.SubElement(root, 'href').text = self._href
        if self._gx_params:
            if self._gx_params.x != 0:
                etree.SubElement(root, 'gx:x').text = str(self._gx_params.x)
            if self._gx_params.y != 0:
                etree.SubElement(root, 'gx:y').text = str(self._gx_params.y)
            if self._gx_params.w != -1:
                etree.SubElement(root, 'gx:w').text = str(self._gx_params.w)
            if self._gx_params.h != -1:
                etree.SubElement(root, 'gx:h').text = str(self._gx_params.h)
        if self._refresh_mode is not None:
            etree.SubElement(root, 'refreshMode').text = self._refresh_mode.value
        if self._refresh_interval is not None:
            etree.SubElement(root, 'refreshInterval').text = f'{self._refresh_interval:0.1f}'
        # if self.view_refresh_mode is not None:
        #     etree.SubElement(root, 'viewRefreshMode').text = self.view_refresh_mode.value
        # if self.view_refresh_time is not None:
        #     etree.SubElement(root, 'viewRefreshTime').text = f'{self.view_refresh_time:0.1f}'
        # if self.view_bound_scale is not None:
        #     etree.SubElement(root, 'viewBoundScale').text = f'{self.view_bound_scale:0.1f}'
        # # TODO: view format is not being handled at all, may need to be corrected
        # if self.view_format is not None:
        #     etree.SubElement(root, 'viewFormat').text = self.view_format
        # if self.http_query:
        #     etree.SubElement(root, 'httpQuery').text = self.http_query

    def __init__(
            self,
            href: Optional[str] = None,
            refresh_mode: Optional[RefreshMode] = None,
            refresh_interval: Optional[float] = None,  # 4.0
            gx_params: Optional[GxParams] = None,  # GxParams(0, 0, -1, -1)
    ):
        Link.__init__(self, href=href, refresh_mode=refresh_mode, refresh_interval=refresh_interval)
        self._gx_params = gx_params



from abc import ABC

from .Object import Object


class Geometry(Object, ABC):
    """A KML 'Geometry', per https://developers.google.com/kml/documentation/kmlreference#geometry. The
    :class:`~pyLiveKML.KML.KMLObjects.Geometry` class is the abstract base class for KML
    :class:`~pyLiveKML.KML.KMLObjects.Object` instances that have an existence as geospatial objects in GEP and that
    are children of :class:`~pyLiveKML.KML.KMLObjects.Placemark` instances.
    """
    def __init__(self):
        Object.__init__(self)
        ABC.__init__(self)



from typing import Optional


class GeoCoordinates:
    """
    The GeoCoordinates type describes a single instance of a Lon-Lat-Alt (LLA) position, as used in
    :class:`~pyLiveKML.KML.KMLObjects.Point`, :class:`~pyLiveKML.KML.KMLObjects.LineString` and
    :class:`~pyLiveKML.KML.KMLObjects.LinearRing` objects. Note that the GeoCoordinates type is *not* explicitly
    referenced by the KML specification; rather, it is a construct of convenience for the pyLiveKML package.

    :param float lon: The longitude.
    :param float lat: The latitude.
    :param Optional[float] alt: The (optional) altitude.
    :var float lon: The longitude, in decimal degrees.
    :var float lat: The latitude, in decimal degrees.
    :var Optional[float] alt: The (optional) altitude, in metres.
    """

    def __init__(
        self,
        lon: float = 0,
        lat: float = 0,
        alt: Optional[float] = None,
    ):
        self.lon: float = lon
        self.lat: float = lat
        self.alt: Optional[float] = alt

    def __str__(self):
        if self.alt is None:
            return f'{self.lon:0.6f},{self.lat:0.6f}'
        else:
            return f'{self.lon:0.6f},{self.lat:0.6f},{self.alt:0.1f}'

    def __repr__(self):
        return self.__str__()












from lxml import etree

from .KML import UnitsEnum, Vec2Type


class Vec2:
    """
    KML field, per the documentation at https://developers.google.com/kml/documentation/kmlreference#kml-fields, that
    is used to define the reference point, handle, or *hotspot* for an icon in GEP.

    :param Vec2Type vec_type: The sub-type of the Vec2 object. Defaults to 'hotSpot'.
    :param float x: The x value of the Vec2; defaults to 0.5.
    :param float y: The y value of the Vec2; defaults to 0.5.
    :param UnitsEnum x_units: The units type of the Vec2 object's x value. Defaults to 'fraction'.
    :param UnitsEnum y_units: The units type of the Vec2 object's y value. Defaults to 'fraction'.

    :var Vec2Type vec_type: The sub-type of the Vec2 object. Defaults to 'hotSpot'.
    :var float x: The x value of the Vec2; defaults to 0.5.
    :var float y: The y value of the Vec2; defaults to 0.5.
    :var UnitsEnum x_units: The units type of the Vec2 object's x value. Defaults to 'fraction'.
    :var UnitsEnum y_units: The units type of the Vec2 object's y value. Defaults to 'fraction'.
    """

    @property
    def xml(self) -> etree.Element:
        """An XML representation of this object.
        """
        root = etree.Element(self.vec_type.value)
        root.attrib['x'] = str(self.x)
        root.attrib['y'] = str(self.y)
        root.attrib['xunits'] = self.x_units.value
        root.attrib['yunits'] = self.y_units.value
        return root

    def __eq__(self, other: 'Vec2') -> bool:
        return False if other is None else \
            isinstance(other, Vec2) and self.name == other.name and self.x == other.x and self.y == other.y \
            and self.x_units == other.x_units and self.y_units == other.y_units

    def __ne__(self, other):
        return not self.__eq__(other)

    def __init__(
            self,
            vec_type: Vec2Type = Vec2Type.HOTSPOT,
            x: float = 0.5,
            y: float = 0.5,
            x_units: UnitsEnum = UnitsEnum.FRACTION,
            y_units: UnitsEnum = UnitsEnum.FRACTION
    ):
        super().__init__()
        self.vec_type = vec_type
        self.x = x
        self.y = y
        self.x_units = x_units
        self.y_units = y_units




from typing import Optional

from lxml import etree

from .KML import KML_UPDATE_CONTAINER_LIMIT_DEFAULT
from .KMLObjects.Container import Container
from .KMLObjects.Folder import Folder


class NetworkLinkControl:
    """The NetworkLinkControl class synchronizes the internal state of a KML object tree rooted in its
    :attr:`container` attribute with GEP.

    :param str target_href: The URI of the KML document that will be synchronized by this
        :class:`~pyLiveKML.KML.NetworkLinkControl`. Note that the specified document *must* already have been created
        in GEP in order to be synchronized.
    :param Optional[Container] container: The KML :class:`~pyLiveKML.KML.KMLObjects.Container` that will be the
        root of the tree of KML :class:`~pyLiveKML.KML.KMLObjects.Feature` instances managed and synchronized by this
        :class:`~pyLiveKML.KML.NetworkLinkControl`.  Note that if no :class:`~pyLiveKML.KML.KMLObjects.Container`, or
        None, is specified then a :class:`~pyLiveKML.KML.KMLObjects.Folder` instance named 'Root' will be instantiated
        for it.
    :param int update_limit: The (approximate) maximum number of Create, Change and/or Delete KML tags that will be
        generated as a result of any single synchronization update.

    :var Container container: The root of the tree of KML :class:`~pyLiveKML.KML.KMLObjects.Feature` instances that this
        :class:`~pyLiveKML.KML.NetworkLinkControl` synchronizes.
    :var str target_href: The URI of the KML document that will be synchronized by this
        :class:`~pyLiveKML.KML.NetworkLinkControl`.
    :var int update_limit: The (approximate) maximum number of KML objects that will be synchronized by any single
        synchronization update.
    """

    def update_kml(self) -> etree.Element:
        """Generate a synchronization update by parsing the :attr:`container`.

        :return: The synchronization update.
        :rtype: etree.Element
        """
        root = etree.Element('NetworkLinkControl')
        update = etree.SubElement(root, 'Update')
        etree.SubElement(update, 'targetHref').text = self.target_href

        for f in self.container.containers:
            f.feature.update_kml(f.container, update)
            for c in f.feature.children:
                c.child.update_kml(c.parent, update)
            for d in f.feature.flush:
                d.delete_kml(update)

        for f in self.container.features:
            if len(update) >= self.container.update_limit:
                break
            f.feature.update_kml(f.container, update)
            for c in f.feature.children:
                c.child.update_kml(c.parent, update)
        return root

    def __init__(
            self,
            target_href: str = '',
            container: Optional[Container] = None,
            update_limit: int = KML_UPDATE_CONTAINER_LIMIT_DEFAULT,
    ):
        # TODO: All of these should be included in the update_kml method, but they're not terribly important ATM.
        # self.min_refresh_period: Optional[float] = None
        # self.max_session_length: Optional[float] = None
        # self.cookie: Optional[str] = None
        # self.message: Optional[str] = None
        # self.link_name: Optional[str] = None
        # self.link_description: Optional[str] = None
        # self.link_snippet: Optional[str] = None
        # self.expires: Optional[datetime] = None
        self.target_href: str = target_href
        self.container: Container = Folder('Root', is_open=True) if container is None else container
        self.update_limit = update_limit



from typing import Optional

from lxml import etree

from ..KML import DisplayMode
from .SubStyle import SubStyle


class BalloonStyle(SubStyle):
    """
    A KML 'BalloonStyle', per https://developers.google.com/kml/documentation/kmlreference#balloonstyle.  Specifies 
    how the description balloon of :class:`~pyLiveKML.KML.KMLObjects.Placemark` objects is drawn.

    :param Optional[str] text: The (optional) text to be displayed in the balloon.
    :param Optional[int] text_color: The (optional) color of the text to be displayed, in 32-bit ABGR format.
    :param Optional[int] bg_color: The (optional) background color of the balloon, in 32-bit ABGR format.
    :param Optional[DisplayMode] display_mode: The (optional) :class:`~pyLiveKML.KML.KML.DisplayMode` of the balloon.
    """

    @property
    def kml_type(self) -> str:
        """Overridden from :attr:`~pyLiveKML.KML.KMLObjects.Object.Object.kml_type` to set the KML tag name to
        'BalloonStyle'"""
        return 'BalloonStyle'

    @property
    def bg_color(self) -> Optional[int]:
        """The background color of the balloon, in 32-bit ABGR format.
        """
        return self._bg_color

    @bg_color.setter
    def bg_color(self, value: Optional[int]):
        if self._bg_color != value:
            self._bg_color = value
            self.field_changed()

    @property
    def text_color(self) -> Optional[int]:
        """The color of the text in the balloon, in 32-bit ABGR format.
        """
        return self._text_color

    @text_color.setter
    def text_color(self, value: Optional[int]):
        if self._text_color != value:
            self._text_color = value
            self.field_changed()

    @property
    def text(self) -> Optional[str]:
        """The text displayed in the balloon.  Note that HTML markup, e.g. tables, is permissible.
        """
        return self._text

    @text.setter
    def text(self, value: Optional[str]):
        if self._text != value:
            self._text = value
            self.field_changed()

    @property
    def display_mode(self) -> Optional[DisplayMode]:
        """The balloon :class:`~pyLiveKML.KML.KML.DisplayMode`.
        """
        return self._display_mode

    @display_mode.setter
    def display_mode(self, value: Optional[DisplayMode]):
        if self._display_mode != value:
            self._display_mode = value
            self.field_changed()

    def build_kml(self, root: etree.Element, with_children=True):
        if self._bg_color is not None:
            etree.SubElement(root, 'bgColor').text = f'{self.bg_color:08x}'
        if self._text_color is not None:
            etree.SubElement(root, 'textColor').text = f'{self.text_color:08x}'
        if self._text is not None:
            etree.SubElement(root, 'text').text = self._text
        if self._display_mode is not None:
            etree.SubElement(root, 'displayMode').text = self._display_mode.value

    def __init__(
            self,
            text: Optional[str] = None,
            text_color: Optional[int] = None,
            bg_color: Optional[int] = None,
            display_mode: Optional[DisplayMode] = None
    ):
        SubStyle.__init__(self)
        self._text = text
        self._bg_color = bg_color
        self._text_color = text_color
        self._display_mode = display_mode

    def __str__(self):
        return f'{self.kml_type}'

    def __repr__(self):
        return self.__str__()



from typing import Optional

from lxml import etree

from ..KML import ListItemType, ItemIconMode
from .SubStyle import SubStyle


class ListStyle(SubStyle):
    """
    A KML 'ListStyle', per https://developers.google.com/kml/documentation/kmlreference#liststyle.  Specifies
    how a :class:`~pyLiveKML.KML.KMLObjects.Feature` is displayed in GEP's user List View.

    :param Optional[ListItemType] list_item_type: The (optional) behaviour model for the list item.
    :param Optional[int] bg_color: The (optional) background color for the list item.
    :param Optional[ItemIconMode] item_icon_state: The (optional) icon state that will be displayed for the list item.
    :param Optional[str] item_icon_href: The (optional) URI for the image will be displayed for the list item.
    """

    @property
    def kml_type(self) -> str:
        """Overridden from :attr:`~pyLiveKML.KML.KMLObjects.Object.Object.kml_type` to set the KML tag name to
        'ListStyle'"""
        return 'ListStyle'

    @property
    def list_item_type(self) -> Optional[ListItemType]:
        """The behaviour model for the list item.
        """
        return self._list_item_type

    @list_item_type.setter
    def list_item_type(self, value: Optional[ListItemType]):
        if self._list_item_type != value:
            self._list_item_type = value
            self.field_changed()

    @property
    def bg_color(self) -> Optional[int]:
        """The background color of the list item, as a 32-bit ABGR color.
        """
        return self._bg_color

    @bg_color.setter
    def bg_color(self, value: Optional[int]):
        if self._bg_color != value:
            self._bg_color = value
            self.field_changed()

    @property
    def item_icon_state(self) -> Optional[ItemIconMode]:
        """The icon state that will be displayed in the GEP user List View for the item.
        """
        return self._item_icon_state

    @item_icon_state.setter
    def item_icon_state(self, value: Optional[ItemIconMode]):
        if self._item_icon_state != value:
            self._item_icon_state = value
            self.field_changed()

    @property
    def item_icon_href(self) -> Optional[str]:
        """The URI for the image that will be displayed in the GEP user List View for the item.
        """
        return self._item_icon_href

    @item_icon_href.setter
    def item_icon_href(self, value: Optional[str]):
        if self._item_icon_href != value:
            self._item_icon_href = value
            self.field_changed()

    def build_kml(self, root: etree.Element, with_children=True):
        if self._list_item_type is not None:
            etree.SubElement(root, 'listItemType').text = self._list_item_type.value
        if self._bg_color is not None:
            etree.SubElement(root, 'bg_color').text = f'{self.bg_color:08x}'
        if self._item_icon_state is not None or self._item_icon_href is not None:
            item_icon = etree.SubElement(root, 'ItemIcon')
            if self._item_icon_state is not None:
                etree.SubElement(item_icon, 'state').text = self._item_icon_state.value
            if self._item_icon_href is not None:
                etree.SubElement(item_icon, 'href').text = self._item_icon_href

    def __init__(
            self,
            list_item_type: Optional[ListItemType] = None,
            bg_color: Optional[int] = None,
            item_icon_state: Optional[ItemIconMode] = None,
            item_icon_href: Optional[str] = None,
    ):
        SubStyle.__init__(self)
        self._list_item_type = list_item_type
        self._bg_color = bg_color
        self._item_icon_state = item_icon_state
        self._item_icon_href = item_icon_href

    def __str__(self):
        return f'{self.kml_type}'

    def __repr__(self):
        return self.__str__()



from typing import Optional
from lxml import etree

from ..GeoCoordinates import GeoCoordinates
from ..KML import AltitudeMode
from .Geometry import Geometry


class Point(Geometry):
    """A Point geometry, as per https://developers.google.com/kml/documentation/kmlreference#point.
    :class:`~pyLiveKML.KML.KMLObjects.Point` objects define a simple geographic location, described a longitude,
    latitude and optional altitude.

    :param GeoCoordinates coordinates: A :class:`~pyLiveKML.KML.GeoCoordinates` object that defines the longitude,
        latitude and optional altitude of the :class:`~pyLiveKML.KML.KMLObjects.Point`.
    :param Optional[bool] extrude: An (optional) flag to indicate whether the :class:`~pyLiveKML.KML.KMLObjects.Point`
        should be shown in GEP connected to the ground with a vertical line.
    :param Optional[bool] altitude_mode: The (optional) :class:`~pyLiveKML.KML.KML.AltitudeMode` that will be applied
        to the :class:`~pyLiveKML.KML.KMLObjects.Point` by GEP.
    """

    @property
    def kml_type(self) -> str:
        """Overridden from :attr:`~pyLiveKML.KML.KMLObjects.Object.Object.kml_type` to set the KML tag name to
        'Point'
        """
        return 'Point'

    @property
    def extrude(self) -> Optional[bool]:
        """True if a vertical line (using the current :class:`~pyLiveKML.KML.KMLObjects.LineStyle`) connects the
        :class:`~pyLiveKML.KML.KMLObjects.Point` to the ground in GEP, False otherwise.  None implies False.
        """
        return self._extrude

    @extrude.setter
    def extrude(self, value: Optional[bool]):
        if self._extrude != value:
            self._extrude = value
            self.field_changed()

    @property
    def altitude_mode(self) -> Optional[AltitudeMode]:
        """An :class:`~pyLiveKML.KML.KML.AltitudeMode` instance that defines how GEP displays the
        :class:`~pyLiveKML.KML.KMLObjects.Point` and treats its' altitude.
        """
        return self._altitude_mode

    @altitude_mode.setter
    def altitude_mode(self, value: Optional[AltitudeMode]):
        if self._altitude_mode != value:
            self._altitude_mode = value
            self.field_changed()

    @property
    def coordinates(self) -> GeoCoordinates:
        """A :class:`~pyLiveKML.KML.GeoCoordinates` object that defines the longitude, latitude and optional altitude
        of the :class:`~pyLiveKML.KML.KMLObjects.Point`.
        """
        return self._coordinates

    @coordinates.setter
    def coordinates(self, value: GeoCoordinates):
        self._coordinates = value
        self.field_changed()

    def build_kml(self, root: etree.Element, with_children=True):
        if self.extrude is not None:
            etree.SubElement(root, 'visibility').text = str(int(self.extrude))
        if self.altitude_mode is not None:
            etree.SubElement(root, 'altitudeMode').text = self.altitude_mode.value
        etree.SubElement(root, 'coordinates').text = self.coordinates.__str__()

    def __init__(
        self,
        coordinates: GeoCoordinates,
        extrude: Optional[bool] = None,
        altitude_mode: Optional[AltitudeMode] = None
    ):
        Geometry.__init__(self)
        self._coordinates: GeoCoordinates = coordinates
        self._extrude = extrude
        self._altitude_mode = altitude_mode

    def __str__(self):
        return f'{self.kml_type}'

    def __repr__(self):
        return self.__str__()



from typing import Optional, Iterator

from lxml import etree

from .BalloonStyle import BalloonStyle
from .IconStyle import IconStyle
from .LabelStyle import LabelStyle
from .LineStyle import LineStyle
from .ListStyle import ListStyle
from .Object import ObjectChild
from .PolyStyle import PolyStyle
from .StyleSelector import StyleSelector


class Style(StyleSelector):
    """A KML 'Style', per https://developers.google.com/kml/documentation/kmlreference#style. A group of
    :class:`~pyLiveKML.KML.KMLObjects.SubStyle` objects that can be referenced by its :attr:`id` or inserted in-line
    into a :class:`~pyLiveKML.KML.KMLObjects.Feature`.

    :param Optional[BalloonStyle] balloon_style: An (optional) :class:`~pyLiveKML.KML.KMLObjects.BalloonStyle` to be
        embedded in this :class:`~pyLiveKML.KML.KMLObjects.Style`.
    :param Optional[IconStyle] icon_style: An (optional) :class:`~pyLiveKML.KML.KMLObjects.IconStyle` to be
        embedded in this :class:`~pyLiveKML.KML.KMLObjects.Style`.
    :param Optional[LabelStyle] label_style: An (optional) :class:`~pyLiveKML.KML.KMLObjects.LabelStyle` to be
        embedded in this :class:`~pyLiveKML.KML.KMLObjects.Style`.
    :param Optional[LineStyle] line_style: An (optional) :class:`~pyLiveKML.KML.KMLObjects.LineStyle` to be
        embedded in this :class:`~pyLiveKML.KML.KMLObjects.Style`.
    :param Optional[ListStyle] list_style: An (optional) :class:`~pyLiveKML.KML.KMLObjects.ListStyle` to be
        embedded in this :class:`~pyLiveKML.KML.KMLObjects.Style`.
    :param Optional[PolyStyle] poly_style: An (optional) :class:`~pyLiveKML.KML.KMLObjects.PolyStyle` to be
        embedded in this :class:`~pyLiveKML.KML.KMLObjects.Style`.
    """

    @property
    def kml_type(self) -> str:
        """Overridden from :attr:`~pyLiveKML.KML.KMLObjects.Object.Object.kml_type` to set the KML tag name to
        'Style'"""
        return 'Style'

    @property
    def children(self) -> Iterator[ObjectChild]:
        """Overridden from :attr:`pyLiveKML.KML.KMLObjects.Object.Object.children` to yield the children of a
        :class:`~pyLiveKML.KML.KMLObjects.Style`, i.e. one or more :class:`~pyLiveKML.KML.KMLObjects.SubStyle`
        instances.
        """
        if self._balloon_style:
            yield ObjectChild(parent=self, child=self._balloon_style)
        if self._icon_style:
            yield ObjectChild(parent=self, child=self._icon_style)
        if self._label_style:
            yield ObjectChild(parent=self, child=self._label_style)
        if self._line_style:
            yield ObjectChild(parent=self, child=self._line_style)
        if self._list_style:
            yield ObjectChild(parent=self, child=self._list_style)
        if self._poly_style:
            yield ObjectChild(parent=self, child=self._poly_style)

    @property
    def balloon_style(self) -> Optional[BalloonStyle]:
        """A :class:`~pyLiveKML.KML.KMLObjects.BalloonStyle` embedded in this :class:`~pyLiveKML.KML.KMLObjects.Style`
        """
        return self._balloon_style

    @property
    def icon_style(self) -> Optional[IconStyle]:
        """An :class:`~pyLiveKML.KML.KMLObjects.IconStyle` embedded in this :class:`~pyLiveKML.KML.KMLObjects.Style`
        """
        return self._icon_style

    @property
    def label_style(self) -> Optional[LabelStyle]:
        """A :class:`~pyLiveKML.KML.KMLObjects.LabelStyle` embedded in this :class:`~pyLiveKML.KML.KMLObjects.Style`
        """
        return self._label_style

    @property
    def line_style(self) -> Optional[LineStyle]:
        """A :class:`~pyLiveKML.KML.KMLObjects.LineStyle` embedded in this :class:`~pyLiveKML.KML.KMLObjects.Style`
        """
        return self._line_style

    @property
    def list_style(self) -> Optional[ListStyle]:
        """A :class:`~pyLiveKML.KML.KMLObjects.ListStyle` embedded in this :class:`~pyLiveKML.KML.KMLObjects.Style`
        """
        return self._list_style

    @property
    def poly_style(self) -> Optional[PolyStyle]:
        """A :class:`~pyLiveKML.KML.KMLObjects.PolyStyle` embedded in this :class:`~pyLiveKML.KML.KMLObjects.Style`
        """
        return self._poly_style

    def build_kml(self, root: etree.Element, with_children=True):
        if with_children:
            if self._balloon_style is not None:
                root.append(self._balloon_style.construct_kml())
            if self._icon_style is not None:
                root.append(self._icon_style.construct_kml())
            if self._label_style is not None:
                root.append(self._label_style.construct_kml())
            if self._line_style is not None:
                root.append(self._line_style.construct_kml())
            if self._list_style is not None:
                root.append(self._list_style.construct_kml())
            if self._poly_style is not None:
                root.append(self._poly_style.construct_kml())

    def __init__(
            self,
            balloon_style: Optional[BalloonStyle] = None,
            icon_style: Optional[IconStyle] = None,
            label_style: Optional[LabelStyle] = None,
            line_style: Optional[LineStyle] = None,
            list_style: Optional[ListStyle] = None,
            poly_style: Optional[PolyStyle] = None,
    ):
        StyleSelector.__init__(self)
        self._balloon_style: Optional[BalloonStyle] = balloon_style
        self._icon_style: Optional[IconStyle] = icon_style
        self._label_style: Optional[LabelStyle] = label_style
        self._line_style: Optional[LineStyle] = line_style
        self._list_style: Optional[ListStyle] = list_style
        self._poly_style: Optional[PolyStyle] = poly_style

    def __str__(self):
        return f'{self.kml_type}'

    def __repr__(self):
        return self.__str__()



from typing import Optional

from lxml import etree

from ..KML import ColorMode
from .ColorStyle import ColorStyle


class LabelStyle(ColorStyle):
    """
    A KML 'LabelStyle', per https://developers.google.com/kml/documentation/kmlreference#iconstyle.  Specifies
    various properties that define how the name of a :class:`~pyLiveKML.KML.KMLObjects.Feature` is drawn in GEP.

    :param Optional[float] scale: The (optional) relative scale of the text.
    :param Optional[int] color: The (optional) color of the text, as a 32-bit ABGR value.
    :param Optional[ColorMode] color_mode: The (optional) :class:`~pyLiveKML.KML.KML.ColorMode` used to color the text;
        either 'NORMAL' or 'RANDOM'.
    """

    @property
    def kml_type(self) -> str:
        """Overridden from :attr:`~pyLiveKML.KML.KMLObjects.Object.Object.kml_type` to set the KML tag name to
        'LabelStyle'"""
        return 'LabelStyle'

    @property
    def scale(self) -> Optional[float]:
        """Relative scale of the text.
        """
        return self._scale

    @scale.setter
    def scale(self, value: Optional[float]):
        if self._scale != value:
            self._scale = value
            self.field_changed()

    def build_kml(self, root: etree.Element, with_children=True):
        if self.color is not None:
            etree.SubElement(root, 'color').text = f'{self.color:08x}'
        if self.color_mode is not None:
            etree.SubElement(root, 'colorMode').text = self.color_mode.value
        if self.scale is not None:
            etree.SubElement(root, 'scale').text = f'{self.scale:0.3f}'

    def __init__(
            self,
            scale: Optional[float] = None,
            color: Optional[int] = None,
            color_mode: Optional[ColorMode] = None,
    ):
        ColorStyle.__init__(self, color=color, color_mode=color_mode)
        self._scale = scale

    def __str__(self):
        return f'{self.kml_type}'

    def __repr__(self):
        return self.__str__()



from abc import ABC
from typing import Optional

from ..KML import ColorMode
from .SubStyle import SubStyle


class ColorStyle(SubStyle, ABC):
    """A KML 'ColorStyle', per https://developers.google.com/kml/documentation/kmlreference#colorstyle.  The
    ColorStyle is the abstract base class for a subset of the specific sub-styles that are optionally included
    in :class:`~pyLiveKML.KML.KMLObjects.Style` objects, and that act to apply a color, typically (but not exclusively)
    to :class:`~pyLiveKML.KML.KMLObjects.Feature` objects.

    :param Optional[int] color: The (optional) color, in ABGR format, that will be applied by GEP if the
        :attr:`color_mode` is :attr:`~pyLiveKML.KML.KML.ColorMode.NORMAL`.
    :param Optional[ColorMode] color_mode: The (optional) :class:`~pyLiveKML.KML.KML.ColorMode` that will be used by
        GEP to determine the displayed color.

    :var Optional[int] color: The (optional) color, in ABGR format, that will be applied by GEP if the
        :attr:`color_mode` is :attr:`~pyLiveKML.KML.KML.ColorMode.NORMAL`.
    :var Optional[ColorMode] color_mode: The (optional) :class:`~pyLiveKML.KML.KML.ColorMode` that will be used by GEP
        to determine the displayed color.
    """
    @property
    def color(self) -> Optional[int]:
        """Color, in 32-bit ABGR format (yes, the order is correct).
        """
        return self._color

    @color.setter
    def color(self, value: Optional[int]):
        val = None if value is None else 0 if value <= 0 else 0xffffffff if value >= 0xffffffff else value
        if self._color != val:
            self._color = val
            self.field_changed()

    @property
    def color_mode(self) -> Optional[ColorMode]:
        """The :class:`~pyLiveKML.KML.KML.ColorMode` that will be used by GEP to determine the displayed color.
        """
        return self._color_mode

    @color_mode.setter
    def color_mode(self, value: Optional[ColorMode]):
        if self._color_mode != value:
            self._color_mode = value
            self.field_changed()

    def __init__(
            self,
            color: Optional[int] = None,
            color_mode: Optional[ColorMode] = None
    ):
        SubStyle.__init__(self)
        ABC.__init__(self)
        self._color = None
        self.color = color
        self._color_mode = color_mode



from typing import Optional, Iterator

from lxml import etree

from .Feature import Feature
from .Geometry import Geometry
from .StyleSelector import StyleSelector
from .Object import ObjectChild


class Placemark(Feature):
    """A KML 'Placemark', per https://developers.google.com/kml/documentation/kmlreference#placemark.
    :class:`~pyLiveKML.KML.KMLObjects.Placemark` objects are containers for a geospatial element to be displayed in GEP.

    :param Geometry geometry: A concrete :class:`~pyLiveKML.KML.KMLObjects.Geometry` instance that will be displayed
        in GEP for this :class:`~pyLiveKML.KML.KMLObjects.Placemark`.
    :param Optional[str] name: The (optional) name that will be displayed in GEP for this
        :class:`~pyLiveKML.KML.KMLObjects.Placemark`.
    :param Optional[bool] visibility: The (optional) initial visibility for this
        :class:`~pyLiveKML.KML.KMLObjects.Placemark` in GEP.
    :param Optional[StyleSelector] inline_style: An (optional) :class:`~pyLiveKML.KML.KMLObjects.StyleSelector` that
        will be applied to this :class:`~pyLiveKML.KML.KMLObjects.Placemark` (only).
    :param Optional[str] style_url: An (optional) style URL, generally a reference to a global
        :class:`~pyLiveKML.KML.KMLObjects.StyleSelector` in a parent of this
        :class:`~pyLiveKML.KML.KMLObjects.Placemark`.
    """

    @property
    def kml_type(self) -> str:
        """Overridden from :attr:`~pyLiveKML.KML.KMLObjects.Object.Object.kml_type` to set the KML tag name to
        'Placemark'
        """
        return 'Placemark'

    @property
    def children(self) -> Iterator[ObjectChild]:
        """Overridden from :attr:`pyLiveKML.KML.KMLObjects.Object.Object.children` to yield the children of a
        :class:`~pyLiveKML.KML.KMLObjects.Placemark`, i.e. a :class:`~pyLiveKML.KML.KMLObjects.Geometry` instance, zero
        or more :class:`~pyLiveKML.KML.KMLObjects.StyleSelector` instances, and any dependent
        :class:`~pyLiveKML.KML.KMLObjects.SubStyle` instances.
        """
        yield ObjectChild(parent=self, child=self.geometry)
        yield from self.geometry.children
        for s in self._styles:
            yield ObjectChild(parent=self, child=s)
            yield from s.children

    @property
    def geometry(self) -> Geometry:
        """The geospatial object, or :class:`~pyLiveKML.KML.KMLObjects.Geometry`, that will be displayed in GEP as
        this :class:`~pyLiveKML.KML.KMLObjects.Placemark`.
        """
        return self._geometry

    def build_kml(self, root: etree.Element, with_children=True):
        if self._name is not None:
            etree.SubElement(root, 'name').text = self.name
        if self._visibility is not None:
            etree.SubElement(root, 'visibility').text = str(int(self.visibility))
        if self._description is not None:
            etree.SubElement(root, 'description').text = self.description
        if self._style_url is not None:
            etree.SubElement(root, 'styleUrl').text = self._style_url
        if with_children:
            for s in self._styles:
                root.append(s.construct_kml())
            root.append(self.geometry.construct_kml())

    def __init__(
            self,
            geometry: Geometry,
            name: Optional[str] = None,
            visibility: Optional[bool] = None,
            inline_style: Optional[StyleSelector] = None,
            style_url: Optional[str] = None
    ):
        Feature.__init__(self, name=name, visibility=visibility, style_url=style_url)
        self._geometry = geometry
        if inline_style:
            self._styles.append(inline_style)



from typing import Optional

from lxml import etree

from ..KML import ColorMode
from .ColorStyle import ColorStyle


class PolyStyle(ColorStyle):
    """
    A KML 'PolyStyle', per https://developers.google.com/kml/documentation/kmlreference#polystyle.  Specifies
    various properties that define how a :class:`~pyLiveKML.KML.KMLObjects.Polygon` is drawn.

    :param Optional[int] color: The (optional) color, as a 32-bit ABGR value, that will be used to :attr:`fill` the
        polygon's area in GEP.
    :param Optional[ColorMode] color_mode: An (optional) :class:`~pyLiveKML.KML.KML.ColorMode` that determines how GEP
        chooses the polygon :attr:`fill` color.
    :param Optional[bool] fill: Optional flag to indicate whether GEP should fill in the polygon with a color.
    :param Optional[bool] outline: Optional flag to indicate whether GEP should draw the polygon's outline.
    """

    @property
    def kml_type(self) -> str:
        """Overridden from :attr:`~pyLiveKML.KML.KMLObjects.Object.Object.kml_type` to set the KML tag name to
        'PolyStyle'"""
        return 'PolyStyle'

    @property
    def fill(self) -> bool:
        """True if the affected :class:`~pyLiveKML.KML.KMLObjects.Polygon` is to be filled in using the :attr:`color`
        property. False if it is not. None implies True.
        """
        return self._fill

    @fill.setter
    def fill(self, value: bool):
        if self._fill != value:
            self._fill = value
            self.field_changed()

    @property
    def outline(self) -> bool:
        """True if the outline of the :class:`~pyLiveKML.KML.KMLObjects.Polygon` is to be drawn using a separate
        :class:`~pyLiveKML.KML.KMLObjects.LineStyle`, specified in the :class:`~pyLiveKML.KML.KMLObjects.StyleSelector`
        that is the parent of this :class:`~pyLiveKML.KML.KMLObjects.PolyStyle` instance. False if it is not. None
        implies True.
        """
        return self._outline

    @outline.setter
    def outline(self, value: bool):
        if self._outline != value:
            self._outline = value
            self.field_changed()

    def build_kml(self, root: etree.Element, with_children=True):
        if self.color is not None:
            etree.SubElement(root, 'color').text = f'{self.color:08x}'
        if self.color_mode is not None:
            etree.SubElement(root, 'colorMode').text = self.color_mode.value
        if self._fill is not None:
            etree.SubElement(root, 'fill').text = str(int(self._fill))
        if self._outline is not None:
            etree.SubElement(root, 'outline').text = str(int(self._outline))

    def __init__(
            self,
            color: Optional[int] = None,
            color_mode: Optional[ColorMode] = None,
            fill: Optional[bool] = None,
            outline: Optional[bool] = None
    ):
        ColorStyle.__init__(self, color, color_mode)
        self._fill: Optional[bool] = fill
        self._outline: Optional[bool] = outline



from typing import Optional, Iterable, Iterator

from lxml import etree

from ..GeoCoordinates import GeoCoordinates
from ..KML import AltitudeMode
from .Geometry import Geometry


class LineString(Geometry):
    """A LineString geometry, as per https://developers.google.com/kml/documentation/kmlreference#linestring.
    :class:`~pyLiveKML.KML.KMLObjects.LineString` objects define an open sequence of points, or
    :class:`~pyLiveKML.KML.GeoCoordinates` such as might make up a track.

    :param Iterable[GeoCoordinates] coordinates: An iterable of :class:`~pyLiveKML.KML.GeoCoordinates` objects, or
        points, that define the :class:`~pyLiveKML.KML.KMLObjects.LineString`. There should be at least two points.
    :param Optional[AltitudeMode] altitude_mode: The (optional) :class:`~pyLiveKML.KML.KML.AltitudeMode` that will be
        applied by GEP to all the points that make up the :class:`~pyLiveKML.KML.KMLObjects.LineString`.
    :param Optional[bool] extrude: An (optional) flag to indicate whether the points that make up the
        :class:`~pyLiveKML.KML.KMLObjects.LineString` should be shown in GEP connected to the ground with vertical
        lines.
    :param Optional[bool] tessellate: An (optional) flag to indicate whether the
        :class:`~pyLiveKML.KML.KMLObjects.LineString` should follow the terrain.
    :param Optional[float] gx_altitude_offset: An (optional) altitude offset (in metres) to be applied to every point
        that makes up the :class:`~pyLiveKML.KML.KMLObjects.LineString`.
    :param Optional[int] gx_draw_order: An (optional) indication of the order in which overlapping
        :class:`~pyLiveKML.KML.KMLObjects.LineString` objects should be drawn.
    """

    @property
    def kml_type(self) -> str:
        """Overridden from :attr:`~pyLiveKML.KML.KMLObjects.Object.Object.kml_type` to set the KML tag name to
        'LineString'"""
        return 'LineString'

    @property
    def gx_altitude_offset(self) -> Optional[float]:
        """An offset, in metres, that is applied to the altitude of all the points
        (:class:`~pyLiveKML.KML.GeoCoordinates`) that define this :class:`~pyLiveKML.KML.KMLObjects.LineString`
        instance.
        """
        return self._gx_altitude_offset

    @gx_altitude_offset.setter
    def gx_altitude_offset(self, value: Optional[float]):
        if self._gx_altitude_offset != value:
            self._gx_altitude_offset = value
            self.field_changed()

    @property
    def extrude(self) -> Optional[bool]:
        """True if a vertical line (using the current :class:`~pyLiveKML.KML.KMLObjects.LineStyle`) connects each of
        the :class:`~pyLiveKML.KML.KMLObjects.LineString` objects' points to the ground in GEP, False otherwise.  None
        implies False.
        """
        return self._extrude

    @extrude.setter
    def extrude(self, value: Optional[bool]):
        if self._extrude != value:
            self._extrude = value
            self.field_changed()

    @property
    def tessellate(self) -> Optional[bool]:
        """True if the boundary line of the :class:`~pyLiveKML.KML.KMLObjects.LineString` follows the terrain in GEP,
        otherwise False.

        :note: The :attr:`altitude_mode` property must be set to CLAMP_TO_GROUND to enable tessellation.
        """
        return self._tessellate

    @tessellate.setter
    def tessellate(self, value: Optional[bool]):
        if self._tessellate != value:
            self._tessellate = value
            self.field_changed()

    @property
    def altitude_mode(self) -> Optional[AltitudeMode]:
        """An :class:`~pyLiveKML.KML.KML.AltitudeMode` instance that defines how GEP displays the
        :class:`~pyLiveKML.KML.GeoCoordinates` objects that make up the  :class:`~pyLiveKML.KML.KMLObjects.LineString`
        and treats their altitudes.
        """
        return self._altitude_mode

    @altitude_mode.setter
    def altitude_mode(self, value: Optional[AltitudeMode]):
        if self._altitude_mode != value:
            self._altitude_mode = value
            self.field_changed()

    @property
    def gx_draw_order(self) -> Optional[int]:
        """An integer that specifies the draw order when multiple :class:`~pyLiveKML.KML.KMLObjects.LineString` objects
        are drawn over the top of one another in GEP. Lower values are drawn first.
        """
        return self._gx_draw_order

    @gx_draw_order.setter
    def gx_draw_order(self, value: Optional[int]):
        if self._gx_draw_order != value:
            self._gx_draw_order = value
            self.field_changed()

    @property
    def coordinates(self) -> Iterator[GeoCoordinates]:
        """A generator to retrieve the :class:`~pyLiveKML.KML.GeoCoordinates` objects that define this
        :class:`~pyLiveKML.KML.KMLObjects.LineString` object.

        :returns: A generator of :class:`~pyLiveKML.KML.GeoCoordinates` objects.
        """
        yield from self._coordinates

    @coordinates.setter
    def coordinates(self, value: Iterable[GeoCoordinates]):
        self._coordinates.clear()
        self._coordinates.extend(value)
        self.field_changed()

    def build_kml(self, root: etree.Element, with_children=True):
        if self._gx_altitude_offset is not None:
            etree.SubElement(root, 'gx:altitudeOffset').text = f'{self._gx_altitude_offset:0.1f}'
        if self._extrude is not None:
            etree.SubElement(root, 'extrude').text = str(int(self._extrude))
        if self._tessellate is not None:
            etree.SubElement(root, 'tessellate').text = str(int(self._tessellate))
        if self._altitude_mode is not None:
            etree.SubElement(root, 'altitudeMode').text = self._altitude_mode.value
        if self._gx_draw_order is not None:
            etree.SubElement(root, 'gx:drawOrder').text = str(self._gx_draw_order)
        if self._coordinates:
            etree.SubElement(root, 'coordinates').text = ' '.join(c.__str__() for c in self._coordinates)

    def __init__(
            self,
            coordinates: Iterable[GeoCoordinates],
            altitude_mode: Optional[AltitudeMode] = None,
            extrude: Optional[bool] = None,
            tessellate: Optional[bool] = None,
            gx_altitude_offset: Optional[float] = None,
            gx_draw_order: Optional[int] = None
    ):
        Geometry.__init__(self)
        self._gx_altitude_offset = gx_altitude_offset
        self._extrude = extrude
        self._tessellate = tessellate
        self._altitude_mode = altitude_mode
        self._gx_draw_order = gx_draw_order
        self._coordinates = list[GeoCoordinates]()
        self._coordinates.extend(coordinates)



from abc import ABC

from .Object import Object


class SubStyle(Object, ABC):
    """A KML 'SubStyle', per https://developers.google.com/kml/documentation/kmlreference. Note that there is no
    description of the :class:`~pyLiveKML.KML.KMLObjects.SubStyle` class in the Google KML documentation, although it
    is included in the inheritance tree at the top of the page.  The :class:`~pyLiveKML.KML.KMLObjects.SubStyle` class
    is the abstract base class for the specific sub-styles that are optionally included as children
    :class:`~pyLiveKML.KML.KMLObjects.Style` objects.
    """

    def __init__(self):
        Object.__init__(self)
        ABC.__init__(self)



import enum

from lxml import etree


KML_UPDATE_CONTAINER_LIMIT_DEFAULT: int = 100
"""The default maximum number of :class:`~pyLiveKML.KML.KMLObjects.Feature` objects that will be included in each 
synchronization update emitted by a :class:`~pyLiveKML.KML.NetworkLinkControl` object.
"""


kml_header: str = '<?xml version="1.0" encoding="UTF-8"?>'
"""The XML tag that opens any XML document, including any KML document.
"""


def kml_tag() -> etree.Element:
    """Construct the opening <kml> tag, with namespaces, for a KML document.

    :return: The <kml> tag, with namespaces, that encloses the contents of a KML document.
    :rtype: etree.Element
    """
    nsmap = {
        'gx': 'http://www.google.com/kml/ext/2.2',
        'kml': 'http://www.opengis.net/kml/2.2',
        'atom': 'http://www.w3.org/2005/Atom',
    }
    attrib = {'xmlns': 'http://www.opengis.net/kml/2.2'}
    return etree.Element('kml', nsmap=nsmap, attrib=attrib)


class AltitudeMode(enum.Enum):
    """Enumeration of options for KML <altitudeMode> tags, generally used in e.g. objects that derive from
    :class:`~pyLiveKML.KML.KMLObjects.Geometry`. Refer to the KML documentation at
    https://developers.google.com/kml/documentation/kmlreference#kml-fields.
    """
    CLAMP_TO_GROUND = 'clampToGround'
    RELATIVE_TO_GROUND = 'relativeToGround'
    ABSOLUTE = 'absolute'


class ColorMode(enum.Enum):
    """Enumeration of options for KML <colorMode> tags, specifically for objects that derive from
    :class:`~pyLiveKML.KML.KMLObjects.ColorStyle`. Refer to the KML documentation at
    https://developers.google.com/kml/documentation/kmlreference#kml-fields.
    """
    NORMAL = 'normal',
    RANDOM = 'random'


class DisplayMode(enum.Enum):
    """Enumeration of options for KML <displayMode> tags, specifically for
    :class:`~pyLiveKML.KML.KMLObjects.BalloonStyle` objects. Refer to the KML documentation at
    https://developers.google.com/kml/documentation/kmlreference#kml-fields.
    """
    DEFAULT = 'default'
    HIDE = 'hide'


class ItemIconMode(enum.Flag):
    """Enumeration of options for KML <ItemIcon><state> tags, specifically for
    :class:`~pyLiveKML.KML.KMLObjects.ListStyle` objects. Refer to the KML documentation at
    https://developers.google.com/kml/documentation/kmlreference#kml-fields.
    """
    OPEN = 'open'
    CLOSED = 'closed'
    ERROR = 'error'
    OPEN_ERROR = 'open error'
    CLOSED_ERROR = 'closed error'
    OPEN_FETCHING0 = 'open fetching0'
    OPEN_FETCHING1 = 'open fetching1'
    OPEN_FETCHING2 = 'open fetching2'
    CLOSED_FETCHING0 = 'closed fetching0'
    CLOSED_FETCHING1 = 'closed fetching1'
    CLOSED_FETCHING2 = 'closed fetching2'


class ListItemType(enum.Enum):
    """Enumeration of options for KML <listItemType> tags, specifically for
    :class:`~pyLiveKML.KML.KMLObjects.ListStyle` objects. Refer to the KML documentation at
    https://developers.google.com/kml/documentation/kmlreference#kml-fields.
    """
    CHECK = 'check'
    CHECK_OFF_ONLY = 'checkOffOnly'
    CHECK_HIDE_CHILDREN = 'checkHideChildren'
    RADIO_FOLDER = 'radioFolder'


class RefreshMode(enum.Enum):
    """Enumeration of options for KML <refreshMode> tags, specifically for
    :class:`~pyLiveKML.KML.KMLObjects.Link` objects. Refer to the KML documentation at
    https://developers.google.com/kml/documentation/kmlreference#kml-fields.
    """
    ON_CHANGE = 'onChange'
    ON_INTERVAL = 'onInterval'
    ON_EXPIRE = 'onExpire'


class StyleState(enum.Enum):
    """Enumeration of options for KML <Pair><key> tags, specifically for
    :class:`~pyLiveKML.KML.KMLObjects.StyleMap` objects. Refer to the KML documentation at
    https://developers.google.com/kml/documentation/kmlreference#kml-fields.
    """
    NORMAL = 'normal'
    HIGHLIGHT = 'highlight'


class UnitsEnum(enum.Enum):
    """Enumeration of options for KML <unitsEnum> tags, specifically for :class:`~pyLiveKML.KML.KMLObjects.Vec2`
    instances in e.g. :class:`~pyLiveKML.KML.KMLObjects.IconStyle` objects. Refer to the KML documentation at
    https://developers.google.com/kml/documentation/kmlreference#kml-fields.
    """
    FRACTION = 'fraction'
    PIXELS = 'pixels'
    INSET_PIXELS = 'insetPixels'


class ViewRefreshMode(enum.Enum):
    """Enumeration of options for KML <viewRefreshMode> tags, specifically for
    :class:`~pyLiveKML.KML.KMLObjects.Link` objects. Refer to the KML documentation at
    https://developers.google.com/kml/documentation/kmlreference#kml-fields.
    """
    NEVER = 'never'
    ON_STOP = 'onStop'
    ON_REQUEST = 'onRequest'
    ON_REGION = 'onRegion'


class Vec2Type(enum.Enum):
    """Enumeration of possible sub-types for KML :class:`~pyLiveKML.KML.KMLObjects.Vec2` objects.  Refer to the KML
    documentation at https://developers.google.com/kml/documentation/kmlreference#kml-fields.
    """
    HOTSPOT = 'hotSpot'


class State(enum.Enum):
    """Enumeration of possible states that objects derived from KML :class:`~pyLiveKML.KML.KMLObjects.Object` may hold.
    The 'State' enumeration is specific to the :mod:`pyLiveKML` package, i.e. it is *not* part of the KML specification.
    """
    IDLE = 0
    CREATING = 1
    CREATED = 2
    CHANGING = 3
    DELETE_CREATED = 4
    DELETE_CHANGED = 5



from typing import Optional, Iterable

from .Feature import Feature
from .Container import Container
from .StyleSelector import StyleSelector


class Folder(Container):
    """A KML 'Folder', per https://developers.google.com/kml/documentation/kmlreference#folder.
    :class:`~pyLiveKML.KML.KMLObjects.Folder` inherits from :class:`~pyLiveKML.KML.KMLObjects.Container` and hence
    :class:`~pyLiveKML.KML.KMLObjects.Folder` objects are containers for :class:`~pyLiveKML.KML.KMLObjects.Feature`
    objects, including other :class:`~pyLiveKML.KML.KMLObjects.Container` instances.

    :param Optional[str] name: The (optional) name for this :class:`~pyLiveKML.KML.KMLObjects.Folder` that will
        be displayed in GEP.
    :param Optional[bool] visibility: The (optional) initial visibility for this
        :class:`~pyLiveKML.KML.KMLObjects.Folder` in GEP.
    :param Optional[bool] is_open: Optional flag to indicate whether the
        :class:`~pyLiveKML.KML.KMLObjects.Folder` will be displayed as 'open' in the GEP user List View.
    :param Optional[str] style_url: An (optional) style URL, generally a reference to a global
        :class:`~pyLiveKML.KML.KMLObjects.StyleSelector` in a parent of this
        :class:`~pyLiveKML.KML.KMLObjects.Folder`.
    :param Optional[Iterable[StyleSelector]] styles: An (optional) iterable of
        :class:`~pyLiveKML.KML.KMLObjects.StyleSelector` instances that will be local to this
        :class:`~pyLiveKML.KML.KMLObjects.Folder`.
    :param Optional[Iterable[Feature]] features: An (optional) iterable of
        :class:`~pyLiveKML.KML.KMLObjects.Feature` instances to be enclosed by this
        :class:`~pyLiveKML.KML.KMLObjects.Folder`.
    """

    @property
    def kml_type(self) -> str:
        """Overridden from :attr:`~pyLiveKML.KML.KMLObjects.Object.Object.kml_type` to set the KML tag name to
        'Folder'"""
        return 'Folder'

    def __init__(
            self,
            name: Optional[str] = None,
            visibility: Optional[bool] = None,
            is_open: Optional[bool] = None,
            style_url: Optional[str] = None,
            styles: Optional[Iterable[StyleSelector]] = None,
            features: Optional[Iterable[Feature]] = None,
    ):
        Container.__init__(
            self,
            name=name,
            visibility=visibility,
            is_open=is_open,
            style_url=style_url,
            styles=styles,
            features=features,
        )



from typing import Optional

from lxml import etree

from ..KML import RefreshMode
from .Object import Object


class Link(Object):
    """A KML 'Link', per https://developers.google.com/kml/documentation/kmlreference#link.
    :class:`~pyLiveKML.KML.KMLObjects.Link` objects are used to specify a KML file to be fetched with a
    :class:`~pyLiveKML.KML.KMLObjects.NetworkLink`; or an image file to be used in a GEP overlay, using the
    :class:`~pyLiveKML.KML.KMLObjects.Icon` class that inherits from :class:`~pyLiveKML.KML.KMLObjects.Link`.  The KML
    specification also uses :class:`~pyLiveKML.KML.KMLObjects.Link` objects to specify the location of Models, but
    Models are not currently implemented in pyLiveKML.

    :note: Only time-based refresh is currently supported by pyLiveKML.

    :param Optional[str] href: An (optional) href for the file that is referenced by the
        :class:`~pyLiveKML.KML.KMLObjects.Link`.
    :param Optional[RefreshMode] refresh_mode: The (optional) refresh mode that will be used for file loading.
    :param Optional[float] refresh_interval: The (optional) refresh interval, in seconds, that will be used for file
        loading.
    """
    @property
    def kml_type(self) -> str:
        """Overridden from :attr:`~pyLiveKML.KML.KMLObjects.Object.Object.kml_type` to set the KML tag name to
        'Link'"""
        return 'Link'

    @property
    def href(self) -> Optional[str]:
        """A URI that specifies the location of the resource that is linked to.
        """
        return self._href

    @href.setter
    def href(self, value: Optional[str]):
        if self._href != value:
            self._href = value
            self.field_changed()

    @property
    def refresh_mode(self) -> Optional[RefreshMode]:
        """The :class:`~pyLiveKML.KML.KML.RefreshMode` that controls when the file specified by the 'href' property is
        retrieved.
        """
        return self._refresh_mode

    @refresh_mode.setter
    def refresh_mode(self, value: Optional[RefreshMode]):
        if self._refresh_mode != value:
            self._refresh_mode = value
            self.field_changed()

    @property
    def refresh_interval(self) -> Optional[float]:
        """The refresh interval, in seconds, that will be used if the :attr:`refresh_mode` property is set to
        :attr:`~pyLiveKML.KML.KML.RefreshMode.ON_INTERVAL`.
        """
        return self._refresh_interval

    @refresh_interval.setter
    def refresh_interval(self, value: Optional[float]):
        if self._refresh_interval != value:
            self._refresh_interval = value
            self.field_changed()

    def build_kml(self, root: etree.Element, with_children=True):
        if self._href:
            etree.SubElement(root, 'href').text = self._href
        if self._refresh_mode is not None:
            etree.SubElement(root, 'refreshMode').text = self._refresh_mode.value
        if self._refresh_interval is not None:
            etree.SubElement(root, 'refreshInterval').text = f'{self._refresh_interval:0.3f}'
        # if self.view_refresh_mode is not None:
        #     etree.SubElement(root, 'viewRefreshMode').text = self.view_refresh_mode.value
        # if self.view_refresh_time is not None:
        #     etree.SubElement(root, 'viewRefreshTime').text = f'{self.view_refresh_time:0.1f}'
        # if self.view_bound_scale is not None:
        #     etree.SubElement(root, 'viewBoundScale').text = f'{self.view_bound_scale:0.1f}'
        # # TODO: view format is not being handled at all, may need to be corrected
        # if self.view_format is not None:
        #     etree.SubElement(root, 'viewFormat').text = self.view_format
        # if self.http_query:
        #     etree.SubElement(root, 'httpQuery').text = self.http_query

    def __init__(
            self,
            href: Optional[str] = None,
            refresh_mode: Optional[RefreshMode] = None,
            refresh_interval: Optional[float] = None,
    ):
        Object.__init__(self)
        self._href = href
        self._refresh_mode = refresh_mode
        self._refresh_interval = refresh_interval
        # self.view_refresh_mode: Optional[ViewRefreshMode] = None
        # self.view_refresh_time: Optional[float] = None
        # self.view_bound_scale: Optional[float] = None
        # self.view_format: Optional[str] = None
        # self.http_query: Optional[str] = None
